SystemOrganization addCategory: #'Morphic-Text Support'!

ParagraphEditor subclass: #TextMorphEditor
	instanceVariableNames: 'morph oldInterval pivotBlock editHistory'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Text Support'!

!TextMorphEditor commentStamp: '<historical>' prior: 0!
This is the ParagraphEditor for TextMorphs.



-----
In the past, BookMorphs had the ability to have each page be on the server as a .sp SqueakPage file.  The index of the book was a .bo file.  In text, Cmd-6 had a LinkTo option that linked to a page by its name, or created a new page of that name.  It assumed the book was on a server with a file per page.  Ted removed that code, and kept a copy on his disk in 'TME-ChngEmphasis.st for .bo .sp'!

!TextMorphEditor methodsFor: 'menu messages' stamp: 'di 4/21/1998 20:30'!
accept
	"Save the current text of the text being edited as the current acceptable version for purposes of canceling.  Allow my morph to take appropriate action"
	morph acceptContents! !

!TextMorphEditor methodsFor: 'multi level undo' stamp: 'sps 7/24/2003 18:56'!
addEditCommand: anEditCommand

	self editHistory rememberCommand: anEditCommand.
"
	Debug dShow: anEditCommand newText.
	Debug dShow: anEditCommand replacedText.
"


! !

!TextMorphEditor methodsFor: 'private' stamp: 'dgd 2/21/2003 22:50'!
againOrSame: bool 
	| bk keys |
	(bk := morph ownerThatIsA: BookMorph) ifNotNil: 
			[(keys := bk valueOfProperty: #tempSearchKey ifAbsent: [nil]) ifNil: 
					["Cmd-f"

					keys := bk valueOfProperty: #searchKey ifAbsent: [nil]	"Cmd-g"]
				ifNotNil: [bk removeProperty: #tempSearchKey].
			keys ifNotNil: 
					[keys notEmpty
						ifTrue: 
							[bk findText: keys.
							^(morph respondsTo: #editView) 
								ifTrue: [morph editView selectionInterval: self selectionInterval]]]].
	super againOrSame: bool.
	(morph respondsTo: #editView) 
		ifTrue: [morph editView selectionInterval: self selectionInterval]! !

!TextMorphEditor methodsFor: 'menu messages' stamp: 'ar 12/17/2001 12:55'!
align
	"Align text according to the next greater alignment value,
	cycling among leftFlush, rightFlush, center, and justified."
	self changeAlignment.
	self recomputeInterval! !

!TextMorphEditor methodsFor: 'binding' stamp: 'ls 7/24/1998 21:06'!
bindingOf: aString
	^model bindingOf: aString! !

!TextMorphEditor methodsFor: 'menu messages' stamp: 'di 10/9/1998 16:55'!
cancel
	"Cancel the changes made so far to this text"
	morph cancelEdits! !

!TextMorphEditor methodsFor: 'editing keys' stamp: 'nk 8/23/2004 17:08'!
changeEmphasis: aStream
	"Change the emphasis of the current selection."
	| retval |
	retval := super changeEmphasis: aStream.
	paragraph composeAll.
	self recomputeInterval.
	morph updateFromParagraph.
	^retval! !

!TextMorphEditor methodsFor: 'attributes' stamp: 'KR 12/5/2005 00:57'!
changeEmphasisOrAlignment
	| aList reply  code align menuList startIndex alignSymbol |
	self flag: #arNote. "Move this up once we get rid of MVC"
	startIndex _ self startIndex.
	aList _ #(normal bold italic narrow underlined struckOut leftFlush centered rightFlush justified).	
	align _ paragraph text alignmentAt: startIndex 
		ifAbsent:[paragraph textStyle alignment].
	alignSymbol _ TextAlignment alignmentSymbol: align.
	code _ paragraph text emphasisAt: startIndex.
	menuList _ WriteStream on: Array new.
	menuList nextPut: (code isZero ifTrue:['<on>'] ifFalse:['<off>']), 'normal' translated.
	menuList nextPutAll: (#(bold italic underlined struckOut) collect:[:emph|
		(code anyMask: (TextEmphasis perform: emph) emphasisCode)
			ifTrue:['<on>', emph asString translated]
			ifFalse:['<off>',emph asString translated]]).
	((paragraph text attributesAt: startIndex forStyle: paragraph textStyle)
		anySatisfy:[:attr| attr isKern and:[attr kern < 0]]) 
			ifTrue:[menuList nextPut:'<on>', 'narrow' translated]
			ifFalse:[menuList nextPut:'<off>', 'narrow' translated].
	menuList nextPutAll: (#(leftFlush centered rightFlush justified) collect:[:type|
		type == alignSymbol
			ifTrue:['<on>',type asString translated]
			ifFalse:['<off>',type asString translated]]).
	aList _ #(normal bold italic underlined struckOut narrow leftFlush centered rightFlush justified).
	reply _ (SelectionMenu labelList: menuList contents lines: #(1 6) selections: aList) startUpWithoutKeyboard.
	reply notNil ifTrue:
		[(#(leftFlush centered rightFlush justified) includes: reply)
			ifTrue:
				[self setAlignment: reply.
				paragraph composeAll.
				self recomputeInterval]
			ifFalse:
				[self setEmphasis: reply.
				paragraph composeAll.
				self recomputeSelection.
				self mvcRedisplay]].
	^ true! !

!TextMorphEditor methodsFor: 'attributes' stamp: 'cds 7/27/2005 22:35'!
changeStyle
	"Let user change styles for the current text pane."
	| aList reply style theStyle menuList |
	self flag: #arNote. "Move this up once we get rid of MVC"
	aList := StrikeFont actualFamilyNames.
	theStyle := paragraph textStyle.
	menuList := aList collect:[:styleName|
		"Hack!! use defaultFont for comparison - we have no name that we could use for compare and the style changes with alignment so they're no longer equal."
		(TextConstants at: styleName) defaultFont == theStyle defaultFont
			ifTrue:['<on>', styleName]
			ifFalse:['<off>',styleName]].
	theStyle = TextStyle default
		ifTrue:[menuList addFirst: '<on>DefaultTextStyle']
		ifFalse:[menuList addFirst: '<off>DefaultTextStyle'].
	aList addFirst: 'DefaultTextStyle'.
	reply := (SelectionMenu labelList: menuList lines: #(1) selections: aList) startUpWithCaption: nil at: ActiveHand position allowKeyboard: false.
	reply ifNotNil:
		[(style := TextStyle named: reply) ifNil: [Beeper beep. ^ true].
		paragraph textStyle: style copy.
		paragraph composeAll.
		self recomputeSelection.
		self mvcRedisplay].
	^ true! !

!TextMorphEditor methodsFor: 'attributes' stamp: 'nk 9/1/2004 14:55'!
changeTextFont
	"Present a menu of available fonts, and if one is chosen, apply it to the current selection.
	If there is no selection, or the selection is empty, apply it to the whole morph."
	| curFont newFont attr startIndex stopIndex |
	startIndex _ self startIndex.
	stopIndex := self stopIndex-1 min: paragraph text size.
	curFont _ (paragraph text fontAt: startIndex withStyle: paragraph textStyle).
	newFont _ StrikeFont fromUser: curFont allowKeyboard: false.
	newFont ifNil:[^self].
	attr _ TextFontReference toFont: newFont.
	stopIndex >= startIndex
		ifTrue: [ paragraph text addAttribute: attr from: startIndex to: stopIndex ]
		ifFalse: [ paragraph text addAttribute: attr from: 1 to: paragraph text size. ].

	paragraph composeAll.
	self recomputeInterval.
! !

!TextMorphEditor methodsFor: 'editing keys' stamp: 'tk 5/7/2001 09:10'!
chooseColor
	| attribute |
	"Make a new Text Color Attribute, let the user pick a color, and return the attribute"

	ColorPickerMorph new
		choseModalityFromPreference;
		sourceHand: morph activeHand;
		target: (attribute _ TextColor color: Color black "default");
		selector: #color:;
		originalColor: Color black;
		putUpFor: morph near: morph fullBoundsInWorld.
	^ attribute
! !

!TextMorphEditor methodsFor: 'multi level undo' stamp: 'sps 2/5/2004 13:26'!
closeTypeIn

	| begin stop rInterval nInterval newText |

	(UndoMessage sends: #noUndoer) ifFalse:[^super closeTypeIn].
	Preferences multipleTextUndo ifTrue: 
		[
		beginTypeInBlock == nil ifFalse:
			[
				begin _ self startOfTyping.
				stop _ self stopIndex.
				rInterval _ (begin "+ UndoMessage argument" 
																to: begin + UndoSelection size - 1).
				nInterval _ begin to: stop - 1.
				(nInterval = rInterval) ifTrue:[ ^super closeTypeIn ].
				newText _ nInterval size > 0
										ifTrue: [ paragraph text 
																copyFrom: nInterval first 
																to: nInterval last ]
										ifFalse: [ self nullText ].
				self addEditCommand: 
				 	(EditCommand
							textMorph: morph
							replacedText: UndoSelection copy
							replacedTextInterval: rInterval
							newText: newText 
							newTextInterval: nInterval)
			].
		].
	
	"Call the super regardless, just to keep the standard undo machine happy"
	^super closeTypeIn
! !

!TextMorphEditor methodsFor: 'controlling' stamp: 'di 4/16/1998 11:33'!
controlInitialize
	"No-op for MVC ParagraphEditor compatibility"! !

!TextMorphEditor methodsFor: 'controlling' stamp: 'di 4/16/1998 11:33'!
controlTerminate
	"No-op for MVC ParagraphEditor compatibility"! !

!TextMorphEditor methodsFor: 'typing support' stamp: 'sw 8/12/2002 01:12'!
dispatchOnCharacter: char with: typeAheadStream
	"Carry out the action associated with this character, if any.
	Type-ahead is passed so some routines can flush or use it."

	((char == Character cr) and: [morph acceptOnCR])
		ifTrue:
			[sensor keyboard.  "Gobble cr -- probably unnecessary."
			self closeTypeIn.
			^ true].

	^ super dispatchOnCharacter: char with: typeAheadStream! !

!TextMorphEditor methodsFor: 'multi level undo' stamp: 'sps 7/24/2003 16:49'!
editHistory
	^morph editHistory
! !

!TextMorphEditor methodsFor: 'menu messages' stamp: 'di 10/5/1998 21:48'!
find
	super find.
	morph installEditorToReplace: self! !

!TextMorphEditor methodsFor: 'displaying' stamp: 'di 4/22/1998 10:21'!
flash
	^ morph flash! !

!TextMorphEditor methodsFor: 'editing keys' stamp: 'tk 5/8/1998 11:08'!
inspectIt: characterStream 
	"Inspect the selection -- invoked via cmd-i.  If there is no current selection, use the current line."

	sensor keyboard.		"flush character"
	self inspectIt.
	^ true! !

!TextMorphEditor methodsFor: 'multi level undo' stamp: 'sps 7/24/2003 12:37'!
isInTypeRun
	^beginTypeInBlock ~~ nil
! !

!TextMorphEditor methodsFor: 'accessing' stamp: 'tk 1/13/1999 07:53'!
morph
	^ morph! !

!TextMorphEditor methodsFor: 'accessing' stamp: 'tk 1/13/1999 07:53'!
morph: aMorph
	"Install a link back to the morph being edited (esp for text links)"
	morph _ aMorph ! !

!TextMorphEditor methodsFor: 'events' stamp: 'th 9/18/2002 11:15'!
mouseDown: evt 
	"An attempt to break up the old processRedButton code into threee phases"
	| clickPoint |

	oldInterval _ self selectionInterval.
	clickPoint _ evt cursorPoint.
	(paragraph clickAt: clickPoint for: model controller: self) ifTrue: [
		pivotBlock _ paragraph characterBlockAtPoint: clickPoint.
		self markBlock: pivotBlock.
		self pointBlock: pivotBlock.
		evt hand releaseKeyboardFocus: self.
		^ self].
	evt shiftPressed
		ifFalse:
			[self closeTypeIn.
			pivotBlock _ paragraph characterBlockAtPoint: clickPoint.
			self markBlock: pivotBlock.
			self pointBlock: pivotBlock.]
		ifTrue:
			[self closeTypeIn.
			self mouseMove: evt].
	self storeSelectionInParagraph! !

!TextMorphEditor methodsFor: 'events' stamp: 'th 9/17/2002 16:45'!
mouseMove: evt 
	"Change the selection in response to moue-down drag"

	pivotBlock ifNil: [^ self].  "Patched during clickAt: repair"
	self pointBlock: (paragraph characterBlockAtPoint: (evt cursorPoint)).
	self storeSelectionInParagraph! !

!TextMorphEditor methodsFor: 'events' stamp: 'th 9/19/2002 18:29'!
mouseUp: evt
	"An attempt to break up the old processRedButton code into threee phases"
	oldInterval ifNil: [^ self].  "Patched during clickAt: repair"
	(self hasCaret 
		and: [oldInterval = self selectionInterval])
		ifTrue: [self selectWord].
	self setEmphasisHere.
	(self isDisjointFrom: oldInterval) ifTrue:
		[otherInterval _ oldInterval].
	self storeSelectionInParagraph! !

!TextMorphEditor methodsFor: 'multi level undo' stamp: 'sps 7/24/2003 17:24'!
multiRedo
	^self multiRedoWithCount: 1
! !

!TextMorphEditor methodsFor: 'multi level undo' stamp: 'sps 7/24/2003 17:28'!
multiRedo: readAheadStream
	sensor keyboard. 	"flush character"
	self closeTypeIn.
	self multiRedoWithCount: 1.
	^true
! !

!TextMorphEditor methodsFor: 'multi level undo' stamp: 'sps 7/24/2003 18:52'!
multiRedoWithCount: count

	| command i lastCommand newSelection saveSelection history |

	count > 0 ifFalse:[ ^self ].

	history _ self editHistory.
	(command _ history nextCommand) isNil
			ifTrue:[ ^self multiUndoError: 'Nothing to redo'].

	saveSelection _ self selectionInterval.
	self deselect.
	i _ 0.
	[i < count] whileTrue: 
		[
		history redo.
		lastCommand _ command.
		((i _ i + 1) < count) ifTrue:
			[
			(command _ history nextCommand) ifNil:[
				self multiUndoError: ('Only ', (i - 1) asString, ' commands to redo.').
				i _ count.]]].

	(newSelection _ lastCommand redoSelectionInterval) isNil
			ifTrue:[ self selectInterval: saveSelection]
			ifFalse:[ self selectInterval: newSelection].

! !

!TextMorphEditor methodsFor: 'multi level undo' stamp: 'sps 7/24/2003 17:24'!
multiUndo
	^self multiUndoWithCount: 1
! !

!TextMorphEditor methodsFor: 'multi level undo' stamp: 'md 8/15/2005 11:18'!
multiUndoError: eString

	Beeper beep
! !

!TextMorphEditor methodsFor: 'multi level undo' stamp: 'sps 7/24/2003 18:53'!
multiUndoWithCount: count

	| command i lastCommand saveSelection newSelection history |

	count > 0 ifFalse:[ ^self ].
	history _ self editHistory.
	(command _ history commandToUndo) 
		ifNil:[ ^self multiUndoError: 'Nothing to undo'].
		
	saveSelection _ self selectionInterval.
	self deselect.
	i _ 0.
	[i < count] whileTrue: 
		[history undo.
		lastCommand _ command.
		((i _ i + 1) < count) ifTrue:
			[(command _ history commandToUndo) ifNil:[
				self multiUndoError: ('Only ', (i - 1) asString, ' commands to undo.').
				i _ count. ]]].

	(newSelection _ lastCommand undoSelectionInterval) isNil
			ifTrue:[ self selectInterval: saveSelection]
			ifFalse:[ self selectInterval: newSelection].

! !

!TextMorphEditor methodsFor: 'menu messages' stamp: 'di 10/2/97 11:38'!
mvcRedisplay
	"Ignore mvcRedisplay requests."! !

!TextMorphEditor methodsFor: 'multi level undo' stamp: 'sps 7/24/2003 13:45'!
noUndoReplace: anInterval with: aText
"This is the zap that multilevel undo uses to do edits. This method bypasses any undo/redo plumbing (in contrast to zapSelection:).  This method is called by an EditCommand (which wants to carry out its paragraph surgery without adding another command to the editHistory)"

	| start stop |
	self deselect.
	start _ (anInterval first max: 1).
	stop _ (anInterval last min: paragraph text size).
	(aText isEmpty and: [stop > start]) ifTrue:
		["If deleting, then set emphasisHere from 1st character of the deletion"
		emphasisHere _ (paragraph text attributesAt: start forStyle: paragraph textStyle)
					select: [:att | att mayBeExtended]].
"Debug dShow: ('zap start->stop: ', (start to: stop) asString)."
	paragraph 
		replaceFrom: start 
		to: stop 
		with: aText 
		displaying: false.  

	UndoMessage sends: #noUndoer . "Keep the normal undo machine happy"
	self userHasEdited  " -- note text now dirty"
! !

!TextMorphEditor methodsFor: 'attributes' stamp: 'th 9/19/2002 18:30'!
offerFontMenu
	"Present a menu of available fonts, and if one is chosen, apply it to the current selection.  
	Use only names of Fonts of this paragraph  "
	| aList reply curFont menuList |
true ifTrue:[^self changeTextFont].
	self flag: #arNote. "Move this up once we get rid of MVC"
	curFont _ (paragraph text fontAt: self startIndex withStyle: paragraph textStyle) fontNameWithPointSize.
	aList _ paragraph textStyle fontNamesWithPointSizes.
	menuList _ aList collect:[:fntName|
		fntName = curFont ifTrue:['<on>',fntName] ifFalse:['<off>',fntName]].
	reply _ (SelectionMenu labelList: menuList selections: aList) startUp.
	reply ~~ nil ifTrue:
		[self replaceSelectionWith:
			(Text string: self selection asString 
				attribute: (TextFontChange fontNumber: (aList indexOf: reply)))] ! !

!TextMorphEditor methodsFor: 'menu commands' stamp: 'sw 12/9/2001 18:55'!
offerMenuFromEsc: characterStream 
	"The escape key was hit while the receiver has the keyboard focus; take action"

	^ ActiveEvent shiftPressed 
		ifTrue:
			[self escapeToDesktop: characterStream]
		ifFalse:
			[self raiseContextMenu: characterStream]! !

!TextMorphEditor methodsFor: 'nonediting/nontyping keys' stamp: 'dgd 10/1/2004 13:23'!
raiseContextMenu: characterStream 
	(morph respondsTo: #editView)
		ifTrue: [morph editView yellowButtonActivity: ActiveEvent shiftPressed]
		ifFalse: [sensor keyboard.
			"consume the character"
			morph yellowButtonActivity: false].
	^ true! !

!TextMorphEditor methodsFor: 'typing support' stamp: 'di 6/14/1998 13:11'!
readKeyboard
	super readKeyboard.
	self storeSelectionInParagraph! !

!TextMorphEditor methodsFor: 'scrolling' stamp: 'di 10/2/97 09:08'!
scrollBy: ignore 
	"Ignore scroll requests."! !

!TextMorphEditor methodsFor: 'current selection' stamp: ''!
select
	"Ignore selection redraw requests."! !

!TextMorphEditor methodsFor: 'current selection' stamp: 'jm 10/28/97 18:31'!
selectAndScroll
	"Ignore scroll requests."! !

!TextMorphEditor methodsFor: 'mvc compatibility' stamp: 'th 9/20/2002 11:26'!
selectAndScrollToTop
	"Scroll until the selection is in the view and then highlight it."

	| lineHeight deltaY rect deltaX |
	lineHeight _ paragraph textStyle lineGrid.
	rect _ morph owner bounds.
	deltaY _ self stopBlock top - rect top.
	deltaY ~= 0 ifTrue: [
		deltaX _ 0.
		deltaY _ (deltaY abs + lineHeight - 1 truncateTo: lineHeight) negated.
		morph editView scrollBy: deltaX@deltaY]! !

!TextMorphEditor methodsFor: 'mvc compatibility' stamp: 'sbw 10/14/1999 16:51'!
selectForTopFrom: start to: stop

	self selectFrom: start to: stop.
	morph editView ifNotNil: [self selectAndScrollToTop]! !

!TextMorphEditor methodsFor: 'new selection' stamp: 'ls 11/10/2002 12:26'!
selectFrom: start to: stop
	"Select the specified characters inclusive."
	self selectInvisiblyFrom: start to: stop.
	self closeTypeIn.
	self storeSelectionInParagraph.
	self setEmphasisHere.
! !

!TextMorphEditor methodsFor: 'accessing' stamp: 'tk 12/18/2001 10:25'!
setSearch: aString
	| bk |
	"Set the FindText and ChangeText to seek aString; except if already seeking aString, leave ChangeText alone so again will repeat last replacement."

	(bk _ morph ownerThatIsA: BookMorph) ifNotNil: [
		bk setProperty: #tempSearchKey 
			toValue: (aString findTokens: Character separators)].


	FindText string = aString
		ifFalse: [FindText _ ChangeText _ aString asText]! !

!TextMorphEditor methodsFor: 'mvc compatibility' stamp: 'th 9/19/2002 18:17'!
storeSelectionInParagraph
	paragraph selectionStart: self startBlock selectionStop: self stopBlock! !

!TextMorphEditor methodsFor: 'editing keys' stamp: 'sw 10/18/1998 10:17'!
tempCommand: characterStream 
	"Experimental.  Triggered by Cmd-t; put trial cmd-key commands here to see how they work, before hanging them on their own cmd accelerators."
	Sensor keyboard.
	morph tempCommand.
	^ true! !

!TextMorphEditor methodsFor: 'attributes' stamp: 'bs 5/20/2009 17:10'!
textAlignment
	"Answer 1..4, representing #leftFlush, #rightFlush, #centered, or #justified"
	^ paragraph text 
		alignmentAt: self pointBlock stringIndex
		ifAbsent: [ paragraph textStyle alignment ]! !

!TextMorphEditor methodsFor: 'attributes' stamp: 'nk 7/3/2003 18:33'!
textAlignmentSymbol
	^#(leftFlush rightFlush centered justified) at: self textAlignment
	! !

!TextMorphEditor methodsFor: 'as yet unclassified' stamp: 'sbw 10/13/1999 22:41'!
totalTextHeight

	^paragraph lines last bottom! !

!TextMorphEditor methodsFor: 'accessing' stamp: 'ar 9/22/2001 16:16'!
transformFrom: owner
	^morph transformFrom: owner! !

!TextMorphEditor methodsFor: 'multi level undo' stamp: 'sps 7/24/2003 13:34'!
undo
	"Reset the state of the paragraph prior to the previous edit.
	 If another ParagraphEditor instance did that edit, UndoInterval is invalid;
	 just recover the contents of the undo-buffer at the start of the paragraph."

	Preferences multipleTextUndo 
		ifTrue: [ ^self multiUndo ]
		ifFalse:[ ^super undo ].
! !

!TextMorphEditor methodsFor: 'scrolling' stamp: ''!
updateMarker
	"Ignore scrollbar redraw requests."
! !

!TextMorphEditor methodsFor: 'accessing' stamp: 'di 4/21/1998 14:11'!
userHasEdited
	"Note that my text is free of user edits."

	morph hasUnacceptedEdits: true! !

!TextMorphEditor methodsFor: 'mvc compatibility' stamp: 'di 4/21/1998 13:26'!
userHasNotEdited
	"Note that my text is free of user edits."

	morph hasUnacceptedEdits: false! !

!TextMorphEditor methodsFor: 'as yet unclassified' stamp: 'sbw 10/13/1999 22:43'!
visibleHeight

	^morph owner bounds height! !

!TextMorphEditor methodsFor: 'menu messages' stamp: 'sps 2/5/2004 13:16'!
zapSelectionWith: aText
	"**overridden to inhibit old-style display"
	| start stop rText rInterval isInTypeRun |
	self deselect.
	start _ self startIndex.
	stop _ self stopIndex.
	(aText isEmpty and: [stop > start]) ifTrue:
		["If deleting, then set emphasisHere from 1st character of the deletion"
		emphasisHere _ (paragraph text attributesAt: start forStyle: paragraph textStyle)
					select: [:att | att mayBeExtended]].
	(start = stop and: [aText size = 0]) ifFalse:
		[
		"===Support for multilevel undo start ==="
		rText _ (paragraph text copyFrom: start to: (stop - 1)).
		rInterval _ start to: (stop - 1).
		isInTypeRun _ self isInTypeRun.
		"===Support for multilevel undo end ==="
		
		paragraph replaceFrom: start to: stop - 1
			with: aText displaying: false.  "** was true in super"
		self computeIntervalFrom: start to: start + aText size - 1.
		UndoInterval _ otherInterval _ self selectionInterval.

		"===Support for multilevel undo start ==="
		 (Preferences multipleTextUndo and: [isInTypeRun not])ifTrue:
				[ self addEditCommand: 
							(EditCommand
									textMorph: morph
									replacedText: rText
									replacedTextInterval: rInterval
									newText: aText 
									newTextInterval: super selectionInterval)].
		"===Support for multilevel undo end ==="].

	self userHasEdited  " -- note text now dirty"! !

Object subclass: #EditCommand
	instanceVariableNames: 'textMorph phase replacedText replacedTextInterval newText newTextInterval lastSelectionInterval'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Text Support'!

!EditCommand commentStamp: '<historical>' prior: 0!
This class handles all paragraph surgery in VI. In general, subclasses of EditCommand should be able to rely on the super class' undo/redo machinery -- only the repeat command needs to be overridden in most cases. This assumes, of course, that the newText, replacedText, newTextInterval, and replacedTextInterval have been set correctly.

When setting the interval, use normal mode style selections, not insert mode selections (see class comment of VIMorphEditor).

Possible useful expressions for doIt or printIt.

Structure:
 instVar1		type -- comment about the purpose of instVar1
 instVar2		type -- comment about the purpose of instVar2

Any further useful comments about the general approach of this implementation.!

!EditCommand class methodsFor: 'instance creation' stamp: 'sps 7/24/2003 17:08'!
textMorph: tm
replacedText: replacedText 
replacedTextInterval: replacedTextInterval
newText: newText 
newTextInterval: newTextInterval


	^(self new)
			textMorph: tm
			replacedText: replacedText 
			replacedTextInterval: replacedTextInterval
			newText: newText 
			newTextInterval: newTextInterval;
			yourself

! !

!EditCommand methodsFor: 'command execution' stamp: 'sps 1/7/2002 21:37'!
doCommand

	^self redoCommand

	! !

!EditCommand methodsFor: 'selection' stamp: 'sps 1/7/2002 19:54'!
doSelectionInterval
	^self redoSelectionInterval! !

!EditCommand methodsFor: 'accessors' stamp: 'sps 1/11/2002 17:12'!
iEditCommand
	^true! !

!EditCommand methodsFor: 'accessors' stamp: 'sps 1/5/2002 14:05'!
lastSelectionInterval
	^lastSelectionInterval! !

!EditCommand methodsFor: 'accessors' stamp: 'sps 1/5/2002 16:36'!
newText
	^newText! !

!EditCommand methodsFor: 'accessors' stamp: 'sps 1/5/2002 16:36'!
newText: aText
	^newText _ aText! !

!EditCommand methodsFor: 'accessors' stamp: 'sps 1/4/2002 22:37'!
newTextInterval
	^newTextInterval! !

!EditCommand methodsFor: 'accessors' stamp: 'sps 1/5/2002 16:36'!
newTextInterval: anInterval
	^newText _ anInterval! !

!EditCommand methodsFor: 'accessors' stamp: 'sps 7/24/2003 17:01'!
pEditor
	^textMorph editor
! !

!EditCommand methodsFor: 'accessors' stamp: 'sps 7/24/2003 13:40'!
phase
	^phase 
! !

!EditCommand methodsFor: 'accessors' stamp: 'sps 7/24/2003 13:40'!
phase: aSymbol
	^phase _ aSymbol
! !

!EditCommand methodsFor: 'accessors' stamp: 'sps 1/11/2002 20:58'!
printOn: aStream

	| |
	aStream 
		nextPutAll: self class name;
		nextPut: $[;
		nextPutAll: ('new: ', newTextInterval asString,' -> "', newText, '", rText: ', replacedTextInterval asString,' -> "', replacedText, '"');
		nextPut: $].! !

!EditCommand methodsFor: 'command execution' stamp: 'sps 7/24/2003 17:04'!
redoCommand

	| |

"Debug dShow: ('rInterval: ', replacedTextInterval asString, '. rText: ', replacedText string, ' nInterval: ', newTextInterval asString, ' nText: ', newText string)."
	self textMorphEditor
		noUndoReplace: replacedTextInterval
		with: newText.

"Debug dShow: ('lastSelInt: ', lastSelectionInterval asString)."
! !

!EditCommand methodsFor: 'selection' stamp: 'sps 7/24/2003 17:34'!
redoSelectionInterval
"Return an interval to be displayed as a subtle selection after undo, or nil"

	^newTextInterval
! !

!EditCommand methodsFor: 'accessors' stamp: 'sps 1/5/2002 16:37'!
replacedText
	^replacedText! !

!EditCommand methodsFor: 'accessors' stamp: 'sps 1/5/2002 16:37'!
replacedText: aText
	^replacedText _ aText! !

!EditCommand methodsFor: 'accessors' stamp: 'sps 1/4/2002 22:30'!
replacedTextInterval
	^replacedTextInterval! !

!EditCommand methodsFor: 'accessors' stamp: 'sps 1/5/2002 16:36'!
replacedTextInterval: anInterval
	^replacedTextInterval _ anInterval! !

!EditCommand methodsFor: 'initialization' stamp: 'sps 7/24/2003 17:01'!
textMorph: tm
replacedText: rText 
replacedTextInterval: rInterval
newText: nText 
newTextInterval: nInterval


	textMorph _ tm.
	replacedText _ rText.
	replacedTextInterval _ rInterval.
	newText _ nText.
	newTextInterval _ nInterval.

! !

!EditCommand methodsFor: 'accessors' stamp: 'sps 7/24/2003 17:04'!
textMorphEditor
	^textMorph editor
! !

!EditCommand methodsFor: 'accessors' stamp: 'sps 7/24/2003 17:05'!
textMorphString
	^textMorph text string 
! !

!EditCommand methodsFor: 'accessors' stamp: 'sps 7/24/2003 17:02'!
textMorphStringSize
	^textMorph text string size
! !

!EditCommand methodsFor: 'command execution' stamp: 'sps 7/24/2003 17:04'!
undoCommand

"Debug dShow: ('new Interval: ', newTextInterval asString, '. rText: ', replacedText string)."

	self textMorphEditor	
		noUndoReplace: newTextInterval
		with: replacedText.
		
	
! !

!EditCommand methodsFor: 'selection' stamp: 'sps 7/24/2003 17:36'!
undoSelection
"Return an interval to be displayed as a selection after undo, or nil"

	^replacedTextInterval first to: (replacedTextInterval first + replacedText size - 1)
! !

!EditCommand methodsFor: 'selection' stamp: 'sps 7/24/2003 17:03'!
undoSelectionInterval
"Return an interval to be displayed as a selection after undo, or nil"

	| i |
	i _ (replacedTextInterval first min: self textMorphStringSize).
	^i to: i - 1
! !

Object subclass: #KeyboardBuffer
	instanceVariableNames: 'event eventUsed'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Text Support'!

!KeyboardBuffer methodsFor: 'as yet unclassified' stamp: 'di 9/30/97 19:53'!
commandKeyPressed
	^ event commandKeyPressed! !

!KeyboardBuffer methodsFor: 'as yet unclassified' stamp: 'di 9/30/97 19:54'!
controlKeyPressed
	^ event controlKeyPressed! !

!KeyboardBuffer methodsFor: 'as yet unclassified' stamp: 'di 9/29/97 12:34'!
flushKeyboard
	eventUsed ifFalse: [^ eventUsed _ true].! !

!KeyboardBuffer methodsFor: 'as yet unclassified' stamp: 'di 9/29/97 12:34'!
keyboard
	eventUsed ifFalse: [eventUsed _ true.  ^ event keyCharacter].
	^ nil! !

!KeyboardBuffer methodsFor: 'as yet unclassified' stamp: 'di 9/29/97 12:34'!
keyboardPeek
	eventUsed ifFalse: [^ event keyCharacter].
	^ nil! !

!KeyboardBuffer methodsFor: 'as yet unclassified' stamp: 'ar 7/19/2002 17:44'!
keyboardPressed
	^eventUsed not! !

!KeyboardBuffer methodsFor: 'as yet unclassified' stamp: 'di 9/30/97 19:54'!
leftShiftDown
	^ event shiftPressed! !

!KeyboardBuffer methodsFor: 'as yet unclassified' stamp: ''!
startingEvent: evt
	event _ evt.
	eventUsed _ false! !

Object subclass: #NewParagraph
	instanceVariableNames: 'text textStyle firstCharacterIndex container lines positionWhenComposed offsetToEnd maxRightX wantsColumnBreaks focused selections'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Text Support'!

!NewParagraph commentStamp: '<historical>' prior: 0!
A Paragraph represents text that has been laid out, or composed, in some container.
	text 		A Text with encoded per-character emphasis.
	textStyle	A TextStyle with font set, line height and horizontal alignment.
	firstCharacterIndex    The starting index in text for this paragraph, allowing
				composition of a long text into a number of containers.
	container	A Rectangle or TextContainer that determines where text can go.
	lines		An Array of TextLines comprising the final layout of the text
				after it has been composed within its container.
	positionWhenComposed   As its name implies.  Allows display at new locations
				without the need to recompose the text.
Lines are ordered vertically.  However, for a given y, there may be several lines in left to right order.  Lines must never be empty, even if text is empty.

Notes on yet another hack - 5 Feb 2001

We really need to clean up #composeLinesFrom:to:delta:into:priorLines:atY:!!!!!!

I added one more habdful of code to correct:

This is an annoying bug that's been around for a couple of years, but I finally figured out how to duplicate the problem, so I figured I'd just report it now.  (It doesn't necessarily have to be fixed for 3.0 if it looks messy, but if it's a simple fix, it would be worth it.)

In Morphic, if you have the following text in a workspace:

This is line 1
This is line 2

**and** you have a return character after line 2, you will normally be able to click the mouse two times below line 2 in order to select all the text.  If you edit line 2 (e.g. so that it reads "line number 2"), you can still select all the text by clicking below the second line.  However, if you edit line 1, you will not be able to select all the text from the bottom in the same way.  Things get messed up such that the last return character seems to be gone.  In this state, if you position the cursor immediately after the 2, and press the right arrow, the cursor jumps to the beginning of line 2... oof. (report by Doug Way)

While I don't have a very deep understanding of the above mentioned method, I was able to determine that text ending in a CR worked better in the editor when the last entry in <lines> had a start of text size + 1 and a stop of text size. I have accordingly added code near the end to ensure this. It seems to have fixed the problem, but we do need to clean this baby up some day. - Bob
!

!NewParagraph methodsFor: 'composition' stamp: 'jm 2/25/2003 16:20'!
OLDcomposeLinesFrom: start to: stop delta: delta into: lineColl priorLines: priorLines atY: startingY 
	"While the section from start to stop has changed, composition may ripple all the way to the end of the text.  However in a rectangular container, if we ever find a line beginning with the same character as before (ie corresponding to delta in the old lines), then we can just copy the old lines from there to the end of the container, with adjusted indices and y-values"

	| charIndex lineY lineHeight scanner line row firstLine lineHeightGuess saveCharIndex hitCR maybeSlide sliding bottom priorIndex priorLine |
	charIndex := start.
	lines := lineColl.
	lineY := startingY.
	lineHeightGuess := textStyle lineGrid.
	maxRightX := container left.
	maybeSlide := stop < text size and: [container isMemberOf: Rectangle].
	sliding := false.
	priorIndex := 1.
	bottom := container bottom.
	scanner := CompositionScanner new text: text textStyle: textStyle.
	firstLine := true.
	[charIndex <= text size and: [lineY + lineHeightGuess <= bottom]] 
		whileTrue: 
			[sliding 
				ifTrue: 
					["Having detected the end of rippling recoposition, we are only sliding old lines"

					priorIndex < priorLines size 
						ifTrue: 
							["Adjust and re-use previously composed line"

							priorIndex := priorIndex + 1.
							priorLine := (priorLines at: priorIndex) slideIndexBy: delta
										andMoveTopTo: lineY.
							lineColl addLast: priorLine.
							lineY := priorLine bottom.
							charIndex := priorLine last + 1]
						ifFalse: 
							["There are no more priorLines to slide."

							sliding := maybeSlide := false]]
				ifFalse: 
					[lineHeight := lineHeightGuess.
					saveCharIndex := charIndex.
					hitCR := false.
					row := container rectanglesAt: lineY height: lineHeight.
					1 to: row size
						do: 
							[:i | 
							(charIndex <= text size and: [hitCR not]) 
								ifTrue: 
									[line := scanner 
												composeFrom: charIndex
												inRectangle: (row at: i)
												firstLine: firstLine
												leftSide: i = 1
												rightSide: i = row size.
									lines addLast: line.
									(text at: line last) = Character cr ifTrue: [hitCR := true].
									lineHeight := lineHeight max: line lineHeight.	"includes font changes"
									charIndex := line last + 1]].

					lineY := lineY + lineHeight.
					row notEmpty 
						ifTrue: 
							[lineY > bottom 
								ifTrue: 
									["Oops -- the line is really too high to fit -- back out"

									charIndex := saveCharIndex.
									row do: [:r | lines removeLast]]
								ifFalse: 
									["It's OK -- the line still fits."

									maxRightX := maxRightX max: scanner rightX.
									1 to: row size - 1
										do: 
											[:i | 
											"Adjust heights across row if necess"

											(lines at: lines size - row size + i) lineHeight: lines last lineHeight
												baseline: lines last baseline].
									charIndex > text size 
										ifTrue: 
											["end of text"

											hitCR 
												ifTrue: 
													["If text ends with CR, add a null line at the end"

													lineY + lineHeightGuess <= container bottom 
														ifTrue: 
															[row := container rectanglesAt: lineY height: lineHeightGuess.
															row notEmpty 
																ifTrue: 
																	[line := (TextLine 
																				start: charIndex
																				stop: charIndex - 1
																				internalSpaces: 0
																				paddingWidth: 0)
																				rectangle: row first;
																				lineHeight: lineHeightGuess baseline: textStyle baseline.
																	lines addLast: line]]].
											lines := lines asArray.
											^maxRightX].
									firstLine := false]].
						
					(maybeSlide and: [charIndex > stop]) 
						ifTrue: 
							["Check whether we are now in sync with previously composed lines"

							
							[priorIndex < priorLines size 
								and: [(priorLines at: priorIndex) first < (charIndex - delta)]] 
									whileTrue: [priorIndex := priorIndex + 1].
							(priorLines at: priorIndex) first = (charIndex - delta) 
								ifTrue: 
									["Yes -- next line will have same start as prior line."

									priorIndex := priorIndex - 1.
									maybeSlide := false.
									sliding := true]
								ifFalse: 
									[priorIndex = priorLines size 
										ifTrue: 
											["Weve reached the end of priorLines,
								so no use to keep looking for lines to slide."

											maybeSlide := false]]]]].
	firstLine 
		ifTrue: 
			["No space in container or empty text"

			line := (TextLine 
						start: start
						stop: start - 1
						internalSpaces: 0
						paddingWidth: 0)
						rectangle: (container topLeft extent: 0 @ lineHeightGuess);
						lineHeight: lineHeightGuess baseline: textStyle baseline.
			lines := Array with: line]
		ifFalse: [self fixLastWithHeight: lineHeightGuess].
	"end of container"
	lines := lines asArray.
	^maxRightX! !

!NewParagraph methodsFor: 'private' stamp: 'di 11/8/97 15:47'!
adjustLineIndicesBy: delta
	firstCharacterIndex _ firstCharacterIndex + delta.
	lines do: [:line | line slide: delta].
! !

!NewParagraph methodsFor: 'private' stamp: 'di 10/26/97 15:57'!
adjustRightX
	| shrink |
	shrink _ container right - maxRightX.
	lines do: [:line | line paddingWidth: (line paddingWidth - shrink)].
	container _ container withRight: maxRightX! !

!NewParagraph methodsFor: 'access' stamp: 'di 11/16/97 09:02'!
adjustedFirstCharacterIndex
	"Return the index in the text where this paragraph WOULD begin if nothing had changed, except the size of the text -- ie if there have only been an insertion of deletion in the preceding morphs"
	offsetToEnd ifNil: [^ -1].
	^ text size - offsetToEnd! !

!NewParagraph methodsFor: 'display' stamp: 'di 8/13/2000 12:27'!
asParagraphForPostscript

	^ self! !

!NewParagraph methodsFor: 'display' stamp: 'lr 7/12/2006 09:27'!
caretWidth
	^ Preferences caretWidth! !

!NewParagraph methodsFor: 'alignment' stamp: 'di 10/25/97 19:26'!
centered 
	textStyle centered! !

!NewParagraph methodsFor: 'selection' stamp: 'ar 4/12/2005 19:53'!
characterBlockAtPoint: aPoint 
	"Answer a CharacterBlock for the character in the text at aPoint."
	| line |
	line _ lines at: (self lineIndexForPoint: aPoint).
	^ ((text string isWideString) ifTrue: [
		MultiCharacterBlockScanner new text: text textStyle: textStyle
	] ifFalse: [CharacterBlockScanner new text: text textStyle: textStyle])
		characterBlockAtPoint: aPoint index: nil
		in: line! !

!NewParagraph methodsFor: 'selection' stamp: 'ar 4/12/2005 19:53'!
characterBlockForIndex: index 
	"Answer a CharacterBlock for the character in text at index."
	| line |
	line _ lines at: (self lineIndexForCharacter: index).
	^ ((text string isWideString) ifTrue: [
		MultiCharacterBlockScanner new text: text textStyle: textStyle
	] ifFalse: [
		CharacterBlockScanner new text: text textStyle: textStyle
	])
		characterBlockAtPoint: nil index: ((index max: line first) min: text size+1)
		in: line! !

!NewParagraph methodsFor: 'editing' stamp: 'mk 5/28/2005 11:15'!
clickAt: clickPoint for: model controller: editor
	"Give sensitive text a chance to fire.  Display flash: (100@100 extent: 100@100)."
	| startBlock action target range boxes box |
	action _ false.
	startBlock _ self characterBlockAtPoint: clickPoint.
	(text attributesAt: startBlock stringIndex forStyle: textStyle) 
		do: [:att | att mayActOnClick ifTrue:
				[(target _ model) ifNil: [target _ editor morph].
				range _ text rangeOf: att startingAt: startBlock stringIndex.
				boxes _ self selectionRectsFrom: (self characterBlockForIndex: range first) 
							to: (self characterBlockForIndex: range last+1).
				box _ boxes detect: [:each | each containsPoint: clickPoint] ifNone: [nil].
				box ifNotNil:
					[ box _ (editor transformFrom: nil) invertBoundsRect: box.
					editor morph allOwnersDo: [ :m | box _ box intersect: (m boundsInWorld) ].
					Utilities awaitMouseUpIn: box
						repeating: []
						ifSucceed: [(att actOnClickFor: target in: self at: clickPoint editor: editor) ifTrue: [action _ true]].
					Cursor currentCursor == Cursor webLink ifTrue:[Cursor normal show].
				]]].
	^ action! !

!NewParagraph methodsFor: 'composition' stamp: 'di 11/8/97 15:31'!
compose: t style: ts from: startingIndex in: textContainer
	text _ t.
	textStyle _ ts.
	firstCharacterIndex _ startingIndex.
	offsetToEnd _ text size - firstCharacterIndex.
	container _ textContainer.
	self composeAll! !

!NewParagraph methodsFor: 'composition' stamp: 'yo 12/20/2002 16:18'!
composeAll
	text string isOctetString ifTrue: [
		^ self composeLinesFrom: firstCharacterIndex to: text size delta: 0
			into: OrderedCollection new priorLines: Array new atY: container top.
	].

	^ self multiComposeLinesFrom: firstCharacterIndex to: text size delta: 0
		into: OrderedCollection new priorLines: Array new atY: container top.
! !

!NewParagraph methodsFor: 'composition' stamp: 'di 11/15/97 09:21'!
composeAllStartingAt: characterIndex
	firstCharacterIndex _ characterIndex.
	offsetToEnd _ text size - firstCharacterIndex.
	self composeAll! !

!NewParagraph methodsFor: 'composition' stamp: 'RAA 5/7/2001 10:58'!
composeLinesFrom: start to: stop delta: delta into: lineColl priorLines: priorLines
	atY: startingY
	"While the section from start to stop has changed, composition may ripple all the way to the end of the text.  However in a rectangular container, if we ever find a line beginning with the same character as before (ie corresponding to delta in the old lines), then we can just copy the old lines from there to the end of the container, with adjusted indices and y-values"

	| newResult |

	newResult _ TextComposer new
		composeLinesFrom: start 
		to: stop 
		delta: delta 
		into: lineColl 
		priorLines: priorLines
		atY: startingY
		textStyle: textStyle 
		text: text 
		container: container
		wantsColumnBreaks: wantsColumnBreaks == true.
	lines _ newResult first asArray.
	maxRightX _ newResult second.
	^maxRightX
! !

!NewParagraph methodsFor: 'composition' stamp: 'di 10/22/97 11:13'!
compositionRectangle
	^ container! !

!NewParagraph methodsFor: 'selection' stamp: 'jm 11/19/97 22:56'!
containsPoint: aPoint
	^ (lines at: (self lineIndexForPoint: aPoint)) rectangle
		containsPoint: aPoint! !

!NewParagraph methodsFor: 'copying' stamp: 'di 5/21/1998 21:45'!
deepCopy
	"Don't want to copy the container (etc) or fonts in the TextStyle."
	| new |
	new _ self copy.
	new textStyle: textStyle copy
		lines: lines copy
		text: text deepCopy.
	^ new! !

!NewParagraph methodsFor: 'selection' stamp: 'di 10/5/1998 12:59'!
defaultCharacterBlock
	^ (CharacterBlock new stringIndex: firstCharacterIndex text: text
			topLeft: lines first topLeft extent: 0 @ 0)
		textLine: lines first! !

!NewParagraph methodsFor: 'fonts-display' stamp: 'nk 3/20/2004 11:13'!
displayOn: aCanvas using: displayScanner at: somePosition
	"Send all visible lines to the displayScanner for display"
	| visibleRectangle offset leftInRun line |
	visibleRectangle _ aCanvas clipRect.
	offset _ (somePosition - positionWhenComposed) truncated.
	leftInRun _ 0.
	(self lineIndexForPoint: visibleRectangle topLeft)
		to: (self lineIndexForPoint: visibleRectangle bottomRight)
		do: [:i | line _ lines at: i.
			self displaySelectionInLine: line on: aCanvas.
			line first <= line last ifTrue:
				[leftInRun _ displayScanner displayLine: line
								offset: offset leftInRun: leftInRun]].
! !

!NewParagraph methodsFor: 'display' stamp: 'bs 5/20/2009 16:25'!
displaySelectionInLine: line on: aCanvas 
	| leftX rightX w caretColor selectionStart selectionStop |
	
	aCanvas isShadowDrawing ifTrue: [ ^self ].	"don't draw selection with shadow"
		
	self selections keysAndValuesDo: [:hand :startStop |
		selectionStart := startStop first.
		selectionStop := startStop second.
	
		(self isSelectionInLine: line start: selectionStart stop: selectionStop) ifTrue: [
			
			leftX := (selectionStart stringIndex < line first) 
							ifTrue: [line left]
							ifFalse: [selectionStart left].
							
			rightX := (selectionStop stringIndex > (line last + 1) or: 
							[selectionStop stringIndex = (line last + 1) and: [selectionStop textLine ~= line]]) 
								ifTrue: [line right]
								ifFalse: [selectionStop left].
		
			selectionStart = selectionStop 
				ifTrue: [
					rightX := rightX + 1.
					w := self caretWidth.
					caretColor := self insertionPointColor.
					1 to: w do: [ :i | 
						"Draw caret triangles at top and bottom"
						aCanvas fillRectangle: ((leftX - w + i - 1) @ (line top + i - 1) 
									extent: ((w - i) * 2 + 3) @ 1)
									color: caretColor.
						
						aCanvas fillRectangle: ((leftX - w + i - 1) @ (line bottom - i) 
									extent: ((w - i) * 2 + 3) @ 1)
									color: caretColor].
						
						aCanvas fillRectangle: (leftX @ line top corner: rightX @ line bottom)
									color: caretColor]
				ifFalse: [
					aCanvas fillRectangle: (leftX @ line top corner: rightX @ line bottom) 
								color: self selectionColor].
			].
	].! !

!NewParagraph methodsFor: 'access' stamp: 'di 10/24/97 17:38'!
extent
	^ container width @ (lines last bottom - lines first top)! !

!NewParagraph methodsFor: 'private' stamp: 'di 4/14/98 13:17'!
fastFindFirstLineSuchThat: lineBlock
	"Perform a binary search of the lines array and return the index
	of the first element for which lineBlock evaluates as true.
	This assumes the condition is one that goes from false to true for
	increasing line numbers (as, eg, yval > somey or start char > somex).
	If lineBlock is not true for any element, return size+1."
	| index low high |
	low _ 1.
	high _ lines size.
	[index _ high + low // 2.
	low > high]
		whileFalse: 
			[(lineBlock value: (lines at: index))
				ifTrue: [high _ index - 1]
				ifFalse: [low _ index + 1]].
	^ low! !

!NewParagraph methodsFor: 'access' stamp: 'di 11/8/97 15:41'!
firstCharacterIndex
	^ firstCharacterIndex! !

!NewParagraph methodsFor: 'composition' stamp: 'RAA 2/25/2001 15:02'!
fixLastWithHeight: lineHeightGuess
"This awful bit is to ensure that if we have scanned all the text and the last character is a CR that there is a null line at the end of lines. Sometimes this was not happening which caused anomalous selections when selecting all the text. This is implemented as a post-composition fixup because I coul;dn't figure out where to put it in the main logic."

	| oldLastLine newRectangle line |

	(text size > 1 and: [text last = Character cr]) ifFalse: [^self].

	oldLastLine _ lines last.
	oldLastLine last - oldLastLine first >= 0 ifFalse: [^self].
	oldLastLine last = text size ifFalse: [^self].

	newRectangle _ oldLastLine left @ oldLastLine bottom 
				extent: 0@(oldLastLine bottom - oldLastLine top).
	"Even though we may be below the bottom of the container,
	it is still necessary to compose the last line for consistency..."

	line _ TextLine start: text size+1 stop: text size internalSpaces: 0 paddingWidth: 0.
	line rectangle: newRectangle.
	line lineHeight: lineHeightGuess baseline: textStyle baseline.
	lines _ lines, (Array with: line).
! !

!NewParagraph methodsFor: 'access' stamp: 'rr 3/22/2004 12:42'!
focused
	focused ifNil: [focused := false].
	^ focused! !

!NewParagraph methodsFor: 'access' stamp: 'rr 3/22/2004 12:41'!
focused: aBoolean
	focused := aBoolean! !

!NewParagraph methodsFor: 'private' stamp: 'RAA 8/30/1998 15:30'!
indentationOfLineIndex: lineIndex ifBlank: aBlock
	"Answer the number of leading tabs in the line at lineIndex.  If there are
	 no visible characters, pass the number of tabs to aBlock and return its value.
	 If the line is word-wrap overflow, back up a line and recur."

	| arrayIndex first last cr |
	cr _ Character cr.
	arrayIndex _ lineIndex.
	[first _ (lines at: arrayIndex) first.
	 first > 1 and: [(text string at: first - 1) ~~ cr]] whileTrue: "word wrap"
		[arrayIndex _ arrayIndex - 1].
	last _ (lines at: arrayIndex) last.
	
	^(text string copyFrom: first to: last) indentationIfBlank: aBlock.
! !

!NewParagraph methodsFor: 'initialize-release' stamp: 'tak 12/21/2004 13:29'!
initialize
	self positionWhenComposed: 0 @ 0! !

!NewParagraph methodsFor: 'display' stamp: 'rr 3/22/2004 19:56'!
insertionPointColor
	self focused ifFalse: [^ Color transparent].
	^ Display depth <= 2
		ifTrue: [Color black]
		ifFalse: [Preferences insertionPointColor]! !

!NewParagraph methodsFor: 'display' stamp: 'bs 5/20/2009 15:27'!
isCursorSelectionInLine: line start: selectionStart stop: selectionStop
	^ (selectionStart = selectionStop) and: [selectionStart textLine = line].! !

!NewParagraph methodsFor: 'display' stamp: 'bs 5/20/2009 16:03'!
isLine: line coveredBy: selectionStart and: selectionStop
	| selectionEndsBelowOrInLine selectionStartsAboveOrInLine |
	
	selectionEndsBelowOrInLine := (selectionStop textLine = line) or: [selectionStop stringIndex > (line last + 1)].
	selectionStartsAboveOrInLine := (selectionStart textLine = line) or: [selectionStart stringIndex < line first].
	
	^	selectionEndsBelowOrInLine & selectionStartsAboveOrInLine.! !

!NewParagraph methodsFor: 'display' stamp: 'bs 5/20/2009 15:39'!
isSelectionInLine: line start: selectionStart stop: selectionStop
	
	^ (self isCursorSelectionInLine: line start: selectionStart stop: selectionStop) or: [
			self isLine: line coveredBy: selectionStart and: selectionStop].! !

!NewParagraph methodsFor: 'alignment' stamp: 'di 10/25/97 19:26'!
justified 
	textStyle justified! !

!NewParagraph methodsFor: 'access' stamp: 'di 10/23/97 21:01'!
lastCharacterIndex
	^ lines last last! !

!NewParagraph methodsFor: 'alignment' stamp: 'di 10/25/97 19:26'!
leftFlush 
	textStyle leftFlush! !

!NewParagraph methodsFor: 'private' stamp: 'di 4/14/98 13:11'!
lineIndexForCharacter: index
	"Answer the index of the line in which to select the character at index."
	^ (self fastFindFirstLineSuchThat: [:line | line first > index]) - 1 max: 1! !

!NewParagraph methodsFor: 'private' stamp: 'di 4/14/98 13:13'!
lineIndexForPoint: aPoint
	"Answer the index of the line in which to select the character nearest to aPoint."
	| i py |
	py _ aPoint y truncated.

	"Find the first line at this y-value"
	i _ (self fastFindFirstLineSuchThat: [:line | line bottom > py]) min: lines size.

	"Now find the first line at this x-value"
	[i < lines size and: [(lines at: i+1) top = (lines at: i) top
				and: [aPoint x >= (lines at: i+1) left]]]
		whileTrue: [i _ i + 1].
	^ i! !

!NewParagraph methodsFor: 'private' stamp: 'RAA 8/30/1998 15:04'!
lineIndexOfCharacterIndex: characterIndex 
	"Answer the line index for a given characterIndex."
	"apparently the selector changed with NewParagraph"

	^self lineIndexForCharacter: characterIndex 
! !

!NewParagraph methodsFor: 'private' stamp: 'di 10/24/97 17:40'!
lines
	^ lines! !

!NewParagraph methodsFor: 'private' stamp: 'edc 6/18/2004 09:10'!
moveBy: delta
	lines do: [:line | line moveBy: delta].
	positionWhenComposed ifNotNil:[
	positionWhenComposed _ positionWhenComposed + delta].
	container _ container translateBy: delta! !

!NewParagraph methodsFor: 'composition' stamp: 'yo 1/3/2003 12:17'!
multiComposeLinesFrom: start to: stop delta: delta into: lineColl priorLines: priorLines
	atY: startingY
	"While the section from start to stop has changed, composition may ripple all the way to the end of the text.  However in a rectangular container, if we ever find a line beginning with the same character as before (ie corresponding to delta in the old lines), then we can just copy the old lines from there to the end of the container, with adjusted indices and y-values"

	| newResult |

	newResult _ MultiTextComposer new
		multiComposeLinesFrom: start 
		to: stop 
		delta: delta 
		into: lineColl 
		priorLines: priorLines
		atY: startingY
		textStyle: textStyle 
		text: text 
		container: container
		wantsColumnBreaks: wantsColumnBreaks == true.
	lines _ newResult first asArray.
	maxRightX _ newResult second.
	"maxRightX printString displayAt: 0@0."
	^maxRightX
! !

!NewParagraph methodsFor: 'access' stamp: 'sbw 10/13/1999 22:31'!
numberOfLines

	^lines size! !

!NewParagraph methodsFor: 'private' stamp: 'di 10/21/97 21:36'!
positionWhenComposed: pos
	positionWhenComposed _ pos! !

!NewParagraph methodsFor: 'composition' stamp: 'yo 12/20/2002 16:18'!
recomposeFrom: start to: stop delta: delta
	"Recompose this paragraph.  The altered portion is between start and stop.
	Recomposition may continue to the end of the text, due to a ripple effect.
	Delta is the amount by which the current text is longer than it was
	when its current lines were composed."
	| startLine newLines |
	"Have to recompose line above in case a word-break was affected."
	startLine _ (self lineIndexForCharacter: start) - 1 max: 1.
	[startLine > 1 and: [(lines at: startLine-1) top = (lines at: startLine) top]]
		whileTrue: [startLine _ startLine - 1].  "Find leftmost of line pieces"
	newLines _ OrderedCollection new: lines size + 1.
	1 to: startLine-1 do: [:i | newLines addLast: (lines at: i)].
	text string isOctetString ifTrue: [
		^ self composeLinesFrom: (lines at: startLine) first to: stop delta: delta
			into: newLines priorLines: lines
			atY: (lines at: startLine) top.
	].
	self multiComposeLinesFrom: (lines at: startLine) first to: stop delta: delta
		into: newLines priorLines: lines
		atY: (lines at: startLine) top.
! !

!NewParagraph methodsFor: 'editing' stamp: 'di 4/28/1999 10:14'!
replaceFrom: start to: stop with: aText displaying: displayBoolean 
	"Edit the text, and then recompose the lines." 
	text replaceFrom: start to: stop with: aText.
	self recomposeFrom: start to: start + aText size - 1 delta: aText size - (stop-start+1)! !

!NewParagraph methodsFor: 'alignment' stamp: 'di 10/25/97 19:26'!
rightFlush 
	textStyle rightFlush! !

!NewParagraph methodsFor: 'display' stamp: 'rr 3/23/2004 19:52'!
selectionColor
	| color |
	Display depth = 1 ifTrue: [^ Color veryLightGray].
	Display depth = 2 ifTrue: [^ Color gray].
	color := Preferences textHighlightColor.
	self focused ifFalse: [color := color alphaMixed: 0.2 with: Color veryVeryLightGray].
	^ color! !

!NewParagraph methodsFor: 'selection' stamp: 'bs 5/20/2009 11:00'!
selectionRects
	"Return an array of rectangles representing the selection region."
	|rects|
	rects := OrderedCollection new.
	self selections keysAndValuesDo: [:hand :startStop | 
		rects addAll: (self selectionRectsFrom: (startStop first) to: (startStop second))].
	^ rects.
	"selectionStart ifNil: [^ Array new].
	^ self selectionRectsFrom: selectionStart to: selectionStop"! !

!NewParagraph methodsFor: 'selection' stamp: 'ls 11/2/2001 23:10'!
selectionRectsFrom: characterBlock1 to: characterBlock2 
	"Return an array of rectangles representing the area between the two character blocks given as arguments."
	| line1 line2 rects cb1 cb2 w |
	characterBlock1 <= characterBlock2
		ifTrue: [cb1 _ characterBlock1.  cb2 _ characterBlock2]
		ifFalse: [cb2 _ characterBlock1.  cb1 _ characterBlock2].
	cb1 = cb2 ifTrue:
		[w _ self caretWidth.
		^ Array with: (cb1 topLeft - (w@0) corner: cb1 bottomLeft + ((w+1)@0))].
	line1 _ self lineIndexForCharacter: cb1 stringIndex.
	line2 _ self lineIndexForCharacter: cb2 stringIndex.
	line1 = line2 ifTrue:
		[^ Array with: (cb1 topLeft corner: cb2 bottomRight)].
	rects _ OrderedCollection new.
	rects addLast: (cb1 topLeft corner: (lines at: line1) bottomRight).
	line1+1 to: line2-1 do: [ :i |
		| line |
		line := lines at: i.
		(line left = rects last left and: [ line right = rects last right ])
			ifTrue: [ "new line has same margins as old one -- merge them, so that the caller gets as few rectangles as possible"
					| lastRect |
					lastRect := rects removeLast.
					rects add: (lastRect bottom: line bottom) ]
			ifFalse: [ "differing margins; cannot merge"
					rects add: line rectangle ] ].

	rects addLast: ((lines at: line2) topLeft corner: cb2 bottomLeft).
	^ rects! !

!NewParagraph methodsFor: 'selection' stamp: 'bs 5/20/2009 11:14'!
selectionStart: startBlock selectionStop: stopBlock
	self selections at: ActiveHand put: {startBlock. stopBlock.}.! !

!NewParagraph methodsFor: 'access' stamp: 'bs 5/20/2009 10:44'!
selections
	^ selections ifNil: [selections := Dictionary new]! !

!NewParagraph methodsFor: 'access' stamp: 'sw 1/13/98 21:31'!
string
	^ text string! !

!NewParagraph methodsFor: 'composition' stamp: 'RAA 5/6/2001 15:09'!
testNewComposeAll
	| newResult |
	self 
		OLDcomposeLinesFrom: firstCharacterIndex 
		to: text size 
		delta: 0
		into: OrderedCollection new 
		priorLines: Array new 
		atY: container top.
	newResult _ TextComposer new
		composeLinesFrom: firstCharacterIndex 
		to: text size 
		delta: 0
		into: OrderedCollection new 
		priorLines: Array new 
		atY: container top
		textStyle: textStyle 
		text: text 
		container: container
		wantsColumnBreaks: false.
	newResult first with: lines do: [ :e1 :e2 |
		e1 longPrintString = e2 longPrintString ifFalse: [self halt].
	].
	newResult second = maxRightX ifFalse: [self halt].
	^{newResult. {lines. maxRightX}}
! !

!NewParagraph methodsFor: 'composition' stamp: 'yo 12/17/2002 14:48'!
testNewComposeAll2
	| newResult |
	newResult _ TextComposer new
		composeLinesFrom: firstCharacterIndex 
		to: text size 
		delta: 0
		into: OrderedCollection new 
		priorLines: Array new 
		atY: container top
		textStyle: textStyle 
		text: text 
		container: container
		wantsColumnBreaks: false.
	^{newResult. {lines. maxRightX}}
! !

!NewParagraph methodsFor: 'composition' stamp: 'yo 12/18/2002 15:00'!
testNewComposeAll3
	| newResult |
	newResult _ TextComposer new
		multiComposeLinesFrom: firstCharacterIndex 
		to: text size 
		delta: 0
		into: OrderedCollection new 
		priorLines: Array new 
		atY: container top
		textStyle: textStyle 
		text: text 
		container: (0@0 extent: 31@60)
		wantsColumnBreaks: false.
	^{newResult. {lines. maxRightX}}
! !

!NewParagraph methodsFor: 'access' stamp: 'di 10/21/97 14:39'!
text
	^ text! !

!NewParagraph methodsFor: 'access' stamp: 'jm 11/19/97 20:27'!
textOwner: ignored  "See TextOnCurve"! !

!NewParagraph methodsFor: 'access' stamp: 'di 10/21/97 14:39'!
textStyle
	^ textStyle! !

!NewParagraph methodsFor: 'access' stamp: 'di 10/23/97 19:33'!
textStyle: aTextStyle 
	"Set the style by which the receiver should display its text."
	textStyle _ aTextStyle! !

!NewParagraph methodsFor: 'private' stamp: 'di 5/21/1998 21:47'!
textStyle: ts lines: l text: t
	"Private -- just a service for deepCopy"
	textStyle _ ts.
	lines _ l.
	text _ t.! !

!NewParagraph methodsFor: 'access' stamp: 'RAA 5/6/2001 15:04'!
wantsColumnBreaks

	^wantsColumnBreaks! !

!NewParagraph methodsFor: 'access' stamp: 'RAA 5/6/2001 15:03'!
wantsColumnBreaks: aBoolean

	wantsColumnBreaks _ aBoolean! !

Object subclass: #StringMorphAttributeScanner
	instanceVariableNames: 'fontNumber textColor emphasis alignment actualFont indent kern'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Text Support'!

!StringMorphAttributeScanner commentStamp: '<historical>' prior: 0!
A StringMorphAttributeScanner provides the interface of a CharacterScanner so that text attributes may be collected from a Text and used elsewhere, like in setting the attributes of a StringMorph.
!

!StringMorphAttributeScanner methodsFor: 'accessing' stamp: 'nk 2/26/2004 13:12'!
actualFont
	"Answer the value of actualFont"

	^ actualFont ifNil: [ TextStyle defaultFont ]! !

!StringMorphAttributeScanner methodsFor: 'scanning' stamp: 'nk 2/26/2004 12:40'!
addEmphasis: anInteger
	"Set the value of emphasis"

	emphasis _ emphasis bitOr: anInteger! !

!StringMorphAttributeScanner methodsFor: 'scanning' stamp: 'nk 2/26/2004 12:41'!
addKern: kernDelta
	"Set the current kern amount."
	kern _ kern + kernDelta! !

!StringMorphAttributeScanner methodsFor: 'accessing' stamp: 'nk 2/26/2004 12:34'!
alignment
	"Answer the value of alignment"

	^ alignment! !

!StringMorphAttributeScanner methodsFor: 'accessing' stamp: 'nk 2/26/2004 12:34'!
emphasis
	"Answer the value of emphasis"

	^ emphasis! !

!StringMorphAttributeScanner methodsFor: 'accessing' stamp: 'nk 2/26/2004 13:14'!
font
	"Answer the value of font"

	^self textStyle fontAt: self fontNumber! !

!StringMorphAttributeScanner methodsFor: 'accessing' stamp: 'nk 2/26/2004 13:11'!
fontNumber
	"Answer the value of font"

	^ fontNumber! !

!StringMorphAttributeScanner methodsFor: 'accessing' stamp: 'nk 2/26/2004 12:34'!
indent
	"Answer the value of indent"

	^ indent! !

!StringMorphAttributeScanner methodsFor: 'scanning' stamp: 'nk 2/26/2004 12:37'!
indentationLevel: anInteger
	"Set the value of indent"

	indent _ anInteger! !

!StringMorphAttributeScanner methodsFor: 'initialize-release' stamp: 'nk 2/26/2004 13:10'!
initialize
	emphasis _ 0.
	indent _ 0.
	kern _ 0.
	fontNumber _ 1.
	actualFont _ TextStyle defaultFont! !

!StringMorphAttributeScanner methodsFor: 'string morph' stamp: 'nk 2/26/2004 13:09'!
initializeFromStringMorph: aStringMorph
	| style |
	actualFont _ aStringMorph font ifNil: [ TextStyle defaultFont ].
	style _ actualFont textStyle.
	emphasis _ actualFont emphasis.
	fontNumber _ (style fontIndexOf: actualFont) ifNil: [ 1 ].
	textColor _ aStringMorph color.
! !

!StringMorphAttributeScanner methodsFor: 'accessing' stamp: 'nk 2/26/2004 12:34'!
kern
	"Answer the value of kern"

	^ kern! !

!StringMorphAttributeScanner methodsFor: 'scanning' stamp: 'nk 2/26/2004 13:09'!
setActualFont: aFont
	"Set the value of actualFont, from a TextFontReference"

	actualFont _ aFont.
	aFont textStyle ifNotNilDo: [ :ts | fontNumber _ ts fontIndexOf: aFont ]! !

!StringMorphAttributeScanner methodsFor: 'scanning' stamp: 'nk 2/26/2004 12:39'!
setAlignment: aSymbol
	"Set the value of alignment"

	alignment _ aSymbol! !

!StringMorphAttributeScanner methodsFor: 'scanning' stamp: 'nk 2/26/2004 13:10'!
setFont: fontNum
	"Set the value of font"

	fontNumber _ fontNum! !

!StringMorphAttributeScanner methodsFor: 'accessing' stamp: 'nk 2/26/2004 12:34'!
textColor
	"Answer the value of textColor"

	^ textColor! !

!StringMorphAttributeScanner methodsFor: 'scanning' stamp: 'nk 2/26/2004 12:34'!
textColor: anObject
	"Set the value of textColor"

	textColor _ anObject! !

!StringMorphAttributeScanner methodsFor: 'accessing' stamp: 'nk 2/26/2004 13:12'!
textStyle
	^self actualFont textStyle ifNil: [ TextStyle default ]! !

Object subclass: #TextComposer
	instanceVariableNames: 'lines maxRightX currentY scanner possibleSlide nowSliding prevIndex prevLines currCharIndex startCharIndex stopCharIndex deltaCharIndex theText theContainer isFirstLine theTextStyle defaultLineHeight actualHeight wantsColumnBreaks'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Text Support'!

!TextComposer class methodsFor: 'as yet unclassified' stamp: 'RAA 5/7/2001 09:31'!
characterForColumnBreak

	^Character value: 12! !

!TextComposer methodsFor: 'as yet unclassified' stamp: 'RAA 5/4/2001 18:09'!
addNullLineForIndex: index
"This awful bit is to ensure that if we have scanned all the text and the last character is a CR that there is a null line at the end of lines. Sometimes this was not happening which caused anomalous selections when selecting all the text. This is implemented as a post-composition fixup because I couldn't figure out where to put it in the main logic."

	| oldLastLine r |

	oldLastLine _ lines last.
	oldLastLine last - oldLastLine first >= 0 ifFalse: [^self].
	oldLastLine last = (index - 1) ifFalse: [^self].

	r _ oldLastLine left @ oldLastLine bottom 
				extent: 0@(oldLastLine bottom - oldLastLine top).
	"Even though we may be below the bottom of the container,
	it is still necessary to compose the last line for consistency..."

	self addNullLineWithIndex: index andRectangle: r.
! !

!TextComposer methodsFor: 'as yet unclassified' stamp: 'RAA 5/5/2001 11:05'!
addNullLineWithIndex: index andRectangle: r

	lines addLast: (
		(
			TextLine 
				start: index 
				stop: index - 1
				internalSpaces: 0 
				paddingWidth: 0
		)
			rectangle: r;
			lineHeight: defaultLineHeight baseline: theTextStyle baseline
	)
! !

!TextComposer methodsFor: 'as yet unclassified' stamp: 'RAA 5/4/2001 11:33'!
checkIfReadyToSlide

	"Check whether we are now in sync with previously composed lines"

	(possibleSlide and: [currCharIndex > stopCharIndex]) ifFalse: [^self].

	[prevIndex < prevLines size
		and: [(prevLines at: prevIndex) first < (currCharIndex - deltaCharIndex)]]
			whileTrue: [prevIndex _ prevIndex + 1].

	(prevLines at: prevIndex) first = (currCharIndex - deltaCharIndex) ifTrue: [
		"Yes -- next line will have same start as prior line."
		prevIndex _ prevIndex - 1.
		possibleSlide _ false.
		nowSliding _ true
	] ifFalse: [
		prevIndex = prevLines size ifTrue: [
			"Weve reached the end of prevLines, so no use to keep looking for lines to slide."
			possibleSlide _ false
		]
	]! !

!TextComposer methodsFor: 'as yet unclassified' stamp: 'RAA 5/6/2001 14:48'!
composeAllLines

	[currCharIndex <= theText size and: 
			[(currentY + defaultLineHeight) <= theContainer bottom]] whileTrue: [

		nowSliding ifTrue: [
			self slideOneLineDown ifNil: [^nil].
		] ifFalse: [
			self composeOneLine ifNil: [^nil].
		]
	].
! !

!TextComposer methodsFor: 'as yet unclassified' stamp: 'th 11/18/2002 19:13'!
composeAllRectangles: rectangles

	| charIndexBeforeLine numberOfLinesBefore reasonForStopping |

	actualHeight _ defaultLineHeight.
	charIndexBeforeLine _ currCharIndex.
	numberOfLinesBefore _ lines size.
	reasonForStopping _ self composeEachRectangleIn: rectangles.

	currentY _ currentY + actualHeight.
	currentY > theContainer bottom ifTrue: [
		"Oops -- the line is really too high to fit -- back out"
		currCharIndex _ charIndexBeforeLine.
		lines size - numberOfLinesBefore timesRepeat: [lines removeLast].
		^self
	].
	
	"It's OK -- the line still fits."
	maxRightX _ maxRightX max: scanner rightX.
	1 to: rectangles size - 1 do: [ :i |
		"Adjust heights across rectangles if necessary"
		(lines at: lines size - rectangles size + i)
			lineHeight: lines last lineHeight
			baseline: lines last baseline
	].
	isFirstLine _ false.
	reasonForStopping == #columnBreak ifTrue: [^nil].
	currCharIndex > theText size ifTrue: [
		^nil		"we are finished composing"
	].
	! !

!TextComposer methodsFor: 'as yet unclassified' stamp: 'ar 12/17/2001 01:59'!
composeEachRectangleIn: rectangles

	| myLine lastChar |

	1 to: rectangles size do: [:i | 
		currCharIndex <= theText size ifFalse: [^false].
		myLine _ scanner 
			composeFrom: currCharIndex 
			inRectangle: (rectangles at: i)				
			firstLine: isFirstLine 
			leftSide: i=1 
			rightSide: i=rectangles size.
		lines addLast: myLine.
		actualHeight _ actualHeight max: myLine lineHeight.  "includes font changes"
		currCharIndex _ myLine last + 1.
		lastChar _ theText at: myLine last.
		lastChar = Character cr ifTrue: [^#cr].
		wantsColumnBreaks ifTrue: [
			lastChar = TextComposer characterForColumnBreak ifTrue: [^#columnBreak].
		].
	].
	^false! !

!TextComposer methodsFor: 'as yet unclassified' stamp: 'RAA 5/7/2001 10:11'!
composeLinesFrom: argStart to: argStop delta: argDelta into: argLinesCollection priorLines: argPriorLines atY: argStartY textStyle: argTextStyle text: argText container: argContainer wantsColumnBreaks: argWantsColumnBreaks

	wantsColumnBreaks _ argWantsColumnBreaks.
	lines _ argLinesCollection.
	theTextStyle _ argTextStyle.
	theText _ argText.
	theContainer _ argContainer.
	deltaCharIndex _ argDelta.
	currCharIndex _ startCharIndex _ argStart.
	stopCharIndex _ argStop.
	prevLines _ argPriorLines.
	currentY _ argStartY.
	defaultLineHeight _ theTextStyle lineGrid.
	maxRightX _ theContainer left.
	possibleSlide _ stopCharIndex < theText size and: [theContainer isMemberOf: Rectangle].
	nowSliding _ false.
	prevIndex _ 1.
	scanner _ CompositionScanner new text: theText textStyle: theTextStyle.
	scanner wantsColumnBreaks: wantsColumnBreaks.
	isFirstLine _ true.
	self composeAllLines.
	isFirstLine ifTrue: ["No space in container or empty text"
		self 
			addNullLineWithIndex: startCharIndex
			andRectangle: (theContainer topLeft extent: 0@defaultLineHeight)
	] ifFalse: [
		self fixupLastLineIfCR
	].
	^{lines asArray. maxRightX}

! !

!TextComposer methodsFor: 'as yet unclassified' stamp: 'dgd 2/22/2003 13:31'!
composeOneLine
	| rectangles |
	rectangles := theContainer rectanglesAt: currentY height: defaultLineHeight.
	rectangles notEmpty 
		ifTrue: [(self composeAllRectangles: rectangles) ifNil: [^nil]]
		ifFalse: [currentY := currentY + defaultLineHeight].
	self checkIfReadyToSlide! !

!TextComposer methodsFor: 'as yet unclassified' stamp: 'RAA 5/4/2001 18:09'!
fixupLastLineIfCR
"This awful bit is to ensure that if we have scanned all the text and the last character is a CR that there is a null line at the end of lines. Sometimes this was not happening which caused anomalous selections when selecting all the text. This is implemented as a post-composition fixup because I couldn't figure out where to put it in the main logic."

	(theText size > 1 and: [theText last = Character cr]) ifFalse: [^self].
	self addNullLineForIndex: theText size + 1.
! !

!TextComposer methodsFor: 'as yet unclassified' stamp: 'RAA 5/6/2001 15:15'!
slideOneLineDown

	| priorLine |

	"Having detected the end of rippling recoposition, we are only sliding old lines"
	prevIndex < prevLines size ifFalse: [
		"There are no more prevLines to slide."
		^nowSliding _ possibleSlide _ false
	].

	"Adjust and re-use previously composed line"
	prevIndex _ prevIndex + 1.
	priorLine _ (prevLines at: prevIndex)
				slideIndexBy: deltaCharIndex andMoveTopTo: currentY.
	lines addLast: priorLine.
	currentY _ priorLine bottom.
	currCharIndex _ priorLine last + 1.
	wantsColumnBreaks ifTrue: [
		priorLine first to: priorLine last do: [ :i |
			(theText at: i) = TextComposer characterForColumnBreak ifTrue: [
				nowSliding _ possibleSlide _ false.
				^nil
			].
		].
	].
! !

Object subclass: #TextContainer
	instanceVariableNames: 'textMorph shadowForm vertProfile minWidth rectangleCache fillsOwner avoidsOcclusions'
	classVariableNames: 'OuterMargin'
	poolDictionaries: ''
	category: 'Morphic-Text Support'!

!TextContainer commentStamp: '<historical>' prior: 0!
A TextContainer models the shape of an ownerMorph, possibly occluded by one or more occludingMorphs, and scans this shape to provide a list of rectangles suitable for layout of text.  It does this by displaying the shadow of the ownerMorph in black, and any occludingMorphs in white, on its shadowForm.  It then scans horizontal strips of appropriate height to find unbroken intervals of black, greater than minWidth in extent.  Conputation of the rectangles is done on demand, and results are cached so that text can be redisplayed without having to recompute the rectangles.!

!TextContainer class methodsFor: 'class initialization' stamp: 'di 11/16/97 09:25'!
initialize    "TextContainer initialize"
	OuterMargin _ 2! !

!TextContainer methodsFor: 'access' stamp: 'di 11/4/97 14:05'!
avoidsOcclusions
	^ avoidsOcclusions ifNil: [false]! !

!TextContainer methodsFor: 'access' stamp: 'di 11/13/97 14:45'!
avoidsOcclusions: aBoolean
	avoidsOcclusions _ aBoolean.
	self releaseCachedState! !

!TextContainer methodsFor: 'container protocol' stamp: 'di 10/27/97 23:09'!
bottom
	"Note we should really check for contiguous pixels here"
	^ (self vertProfile findLast: [:count | count >= minWidth])
		+ shadowForm offset y! !

!TextContainer methodsFor: 'private' stamp: 'ar 10/26/2000 20:04'!
bounds
	| bounds theText |
	self fillsOwner ifFalse: [^ textMorph textBounds].
	theText _ textMorph.
	bounds _ theText owner innerBounds.
	bounds _ bounds insetBy: (textMorph valueOfProperty: #margins ifAbsent: [1@1]).
	theText owner submorphsBehind: theText do:
		[:m | bounds _ bounds merge: m fullBounds].
	^ bounds! !

!TextContainer methodsFor: 'private' stamp: 'ar 10/26/2000 20:05'!
computeShadow
	| canvas back bounds theText |
	bounds _ self bounds.
	theText _ textMorph.
	canvas _ (Display defaultCanvasClass extent: bounds extent depth: 1)
			shadowColor: Color black.
	canvas translateBy: bounds topLeft negated during:[:tempCanvas|
		self fillsOwner
			ifTrue: [tempCanvas fullDrawMorph: (theText owner copyWithoutSubmorph: theText)]
			ifFalse: [tempCanvas fillRectangle: textMorph bounds color: Color black].
		self avoidsOcclusions ifTrue:
			[back _ tempCanvas form deepCopy.
			tempCanvas form fillWhite.
			theText owner submorphsInFrontOf: theText do:
				[:m | (textMorph isLinkedTo: m)
					ifTrue: []
					ifFalse: [tempCanvas fullDrawMorph: m]].
			back displayOn: tempCanvas form at: 0@0 rule: Form reverse].
	].
	shadowForm _ canvas form offset: bounds topLeft.
	vertProfile _ shadowForm  yTallyPixelValue: 1 orNot: false.
	rectangleCache _ Dictionary new.
	^ shadowForm! !

!TextContainer methodsFor: 'access' stamp: 'di 11/4/97 14:05'!
fillsOwner
	^ fillsOwner ifNil: [true]! !

!TextContainer methodsFor: 'access' stamp: 'di 11/13/97 14:45'!
fillsOwner: aBoolean
	fillsOwner _ aBoolean.
	self releaseCachedState! !

!TextContainer methodsFor: 'private' stamp: 'di 11/4/97 14:06'!
for: aTextMorph minWidth: wid
	textMorph _ aTextMorph.
	minWidth _ wid.
	fillsOwner _ true.
	avoidsOcclusions _ false.! !

!TextContainer methodsFor: 'container protocol' stamp: 'di 10/28/97 18:33'!
left 
	^ textMorph owner left! !

!TextContainer methodsFor: 'access' stamp: 'yo 1/3/2003 12:21'!
paragraphClass
	^ MultiNewParagraph! !

!TextContainer methodsFor: 'container protocol' stamp: 'dgd 2/22/2003 19:06'!
rectanglesAt: lineY height: lineHeight 
	"Return a list of rectangles that are at least minWidth wide
	in the specified horizontal strip of the shadowForm.
	Cache the results for later retrieval if the owner does not change."

	| hProfile rects thisWidth thisX count pair outerWidth lineRect lineForm |
	pair := Array with: lineY with: lineHeight.
	rects := rectangleCache at: pair ifAbsent: [nil].
	rects ifNotNil: [^rects].
	outerWidth := minWidth + (2 * OuterMargin).
	self shadowForm.	"Compute the shape"
	lineRect := 0 @ (lineY - shadowForm offset y) 
				extent: shadowForm width @ lineHeight.
	lineForm := shadowForm copy: lineRect.

	"Check for a full line -- frequent case"
	(lineForm tallyPixelValues second) = lineRect area 
		ifTrue: 
			[rects := Array with: (shadowForm offset x @ lineY extent: lineRect extent)]
		ifFalse: 
			["No such luck -- scan the horizontal profile for segments of minWidth"

			hProfile := lineForm xTallyPixelValue: 1 orNot: false.
			rects := OrderedCollection new.
			thisWidth := 0.
			thisX := 0.
			1 to: hProfile size
				do: 
					[:i | 
					count := hProfile at: i.
					count >= lineHeight 
						ifTrue: [thisWidth := thisWidth + 1]
						ifFalse: 
							[thisWidth >= outerWidth 
								ifTrue: 
									[rects addLast: ((thisX + shadowForm offset x) @ lineY 
												extent: thisWidth @ lineHeight)].
							thisWidth := 0.
							thisX := i]].
			thisWidth >= outerWidth 
				ifTrue: 
					[rects addLast: ((thisX + shadowForm offset x) @ lineY 
								extent: thisWidth @ lineHeight)]].
	rects := rects collect: [:r | r insetBy: OuterMargin @ 0].
	rectangleCache at: pair put: rects.
	^rects! !

!TextContainer methodsFor: 'access' stamp: 'di 11/16/97 09:39'!
releaseCachedState

	shadowForm _ nil.
	vertProfile _ nil.
	rectangleCache _ Dictionary new.
! !

!TextContainer methodsFor: 'private' stamp: 'di 10/27/97 23:09'!
shadowForm
	shadowForm ifNil: [self computeShadow].
	^ shadowForm! !

!TextContainer methodsFor: 'access' stamp: 'tk 8/31/2000 14:50'!
textMorph
	^ textMorph! !

!TextContainer methodsFor: 'container protocol' stamp: 'di 11/16/97 09:33'!
top
	"Note we should really check for contiguous pixels here"
	| outerWidth |
	outerWidth _ minWidth + (2*OuterMargin).
	^ (self vertProfile findFirst: [:count | count >= outerWidth]) - 1
		+ shadowForm offset y! !

!TextContainer methodsFor: 'container protocol' stamp: 'di 10/28/97 18:33'!
topLeft  "for compatibility"
	^ textMorph owner topLeft! !

!TextContainer methodsFor: 'container protocol' stamp: 'di 11/7/97 12:01'!
translateBy: delta
	self releaseCachedState! !

!TextContainer methodsFor: 'private' stamp: 'di 10/27/97 23:08'!
vertProfile
	vertProfile ifNil: [self computeShadow].
	^ vertProfile! !

!TextContainer methodsFor: 'container protocol' stamp: 'di 10/28/97 18:32'!
width  "for compatibility"
	^ textMorph owner width! !

Object subclass: #TextLine
	instanceVariableNames: 'left right top bottom firstIndex lastIndex internalSpaces paddingWidth baseline leftMargin'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Text Support'!

!TextLine commentStamp: '<historical>' prior: 0!
A TextLine embodies the layout of a line of composed text.
	left right top bottom		The full line rectangle
	firstIndex lastIndex		Starting and stopping indices in the full text
	internalSpaces		Number of spaces to share paddingWidth
	paddingWidth		Number of pixels of extra space in full line
	baseline				Distance of baseline below the top of the line
	leftMargin			Left margin due to paragraph indentation
TextLine's rather verbose message protocol is required for compatibility with the old CharacterScanners.!

!TextLine class methodsFor: 'instance creation' stamp: 'di 10/20/97 23:08'!
start: startInteger stop: stopInteger internalSpaces: spacesInteger paddingWidth: padWidthInteger
	"Answer an instance of me with the arguments as the start, stop points, 
	number of spaces in the line, and width of the padding."
	| line |
	line _ self new firstIndex: startInteger lastIndex: stopInteger.
	^ line internalSpaces: spacesInteger paddingWidth: padWidthInteger! !

!TextLine methodsFor: 'comparing' stamp: 'di 10/20/97 23:24'!
= line

	self species = line species
		ifTrue: [^((firstIndex = line first and: [lastIndex = line last])
				and: [internalSpaces = line internalSpaces])
				and: [paddingWidth = line paddingWidth]]
		ifFalse: [^false]! !

!TextLine methodsFor: 'accessing' stamp: ''!
baseline
	^ baseline! !

!TextLine methodsFor: 'accessing' stamp: 'di 10/23/97 19:58'!
bottom
	^ bottom! !

!TextLine methodsFor: 'accessing' stamp: 'di 10/23/97 19:58'!
bottomRight
	^ right@bottom! !

!TextLine methodsFor: 'accessing' stamp: 'di 10/21/97 20:12'!
first
	^ firstIndex! !

!TextLine methodsFor: 'private' stamp: 'di 10/20/97 23:08'!
firstIndex: firstInteger lastIndex: lastInteger
	firstIndex _ firstInteger.
	lastIndex _ lastInteger! !

!TextLine methodsFor: 'comparing' stamp: 'ar 9/9/2003 22:03'!
hash
	"#hash is re-implemented because #= is re-implemented"
	^firstIndex hash bitXor: lastIndex hash! !

!TextLine methodsFor: 'accessing' stamp: ''!
internalSpaces
	"Answer the number of spaces in the line."

	^internalSpaces! !

!TextLine methodsFor: 'accessing' stamp: ''!
internalSpaces: spacesInteger 
	"Set the number of spaces in the line to be spacesInteger."

	internalSpaces _ spacesInteger! !

!TextLine methodsFor: 'private' stamp: ''!
internalSpaces: spacesInteger paddingWidth: padWidthInteger

	internalSpaces _ spacesInteger.
	paddingWidth _ padWidthInteger! !

!TextLine methodsFor: 'scanning' stamp: ''!
justifiedPadFor: spaceIndex 
	"Compute the width of pad for a given space in a line of justified text."

	| pad |
	internalSpaces = 0 ifTrue: [^0].
	pad _ paddingWidth // internalSpaces.
	spaceIndex <= (paddingWidth \\ internalSpaces)
		ifTrue: [^pad + 1]
		ifFalse: [^pad]! !

!TextLine methodsFor: 'scanning' stamp: ''!
justifiedTabDeltaFor: spaceIndex 
	"Compute the delta for a tab in a line of justified text, so tab falls 
	somewhere plausible when line is justified."

	| pad extraPad |
	internalSpaces = 0 ifTrue: [^0].
	pad _ paddingWidth // internalSpaces.
	extraPad _ paddingWidth \\ internalSpaces.
	spaceIndex <= extraPad
		ifTrue: [^spaceIndex * (pad + 1)]
		ifFalse: [^extraPad * (pad + 1) + (spaceIndex - extraPad * pad)]! !

!TextLine methodsFor: 'accessing' stamp: 'di 10/21/97 20:14'!
last
	^ lastIndex! !

!TextLine methodsFor: 'accessing' stamp: 'di 10/23/97 19:58'!
left
	^ left! !

!TextLine methodsFor: 'accessing' stamp: 'di 10/21/97 20:42'!
leftMargin
	"This has to get fixed -- store during composition"
	^ self left! !

!TextLine methodsFor: 'accessing' stamp: 'hmm 2/9/2001 11:58'!
leftMargin: lm
	left _ lm! !

!TextLine methodsFor: 'accessing' stamp: 'di 10/26/97 16:03'!
leftMarginForAlignment: alignmentCode
	alignmentCode = 1 ifTrue: [^ self left + paddingWidth].  "right flush"
	alignmentCode = 2 ifTrue: [^ self left + (paddingWidth//2)].  "centered"
	^ self left  "leftFlush and justified"! !

!TextLine methodsFor: 'accessing' stamp: 'di 10/23/97 19:58'!
lineHeight
	^ bottom - top! !

!TextLine methodsFor: 'private' stamp: 'di 10/23/97 19:57'!
lineHeight: height baseline: ascent
	bottom _ top + height.
	baseline _ ascent! !

!TextLine methodsFor: 'updating' stamp: 'di 11/7/97 08:32'!
moveBy: delta 
	"Move my rectangle by the given delta"
	left _ left + delta x.
	right _ right + delta x.
	top _ top + delta y.
	bottom _ bottom + delta y.
! !

!TextLine methodsFor: 'accessing' stamp: ''!
paddingWidth
	"Answer the amount of space to be added to the font."

	^paddingWidth! !

!TextLine methodsFor: 'accessing' stamp: ''!
paddingWidth: padWidthInteger 
	"Set the amount of space to be added to the font to be padWidthInteger."

	paddingWidth _ padWidthInteger! !

!TextLine methodsFor: 'printing' stamp: 'di 10/23/97 23:19'!
printOn: aStream
	super printOn: aStream.
	aStream space; print: firstIndex; nextPutAll: ' to: '; print: lastIndex! !

!TextLine methodsFor: 'accessing' stamp: 'di 10/23/97 20:00'!
rectangle
	^ self topLeft corner: self bottomRight! !

!TextLine methodsFor: 'accessing' stamp: 'hmm 2/9/2001 11:58'!
rectangle: lineRectangle
	left _ lineRectangle left.
	right _ lineRectangle right.
	top _ lineRectangle top.
	bottom _ lineRectangle bottom! !

!TextLine methodsFor: 'accessing' stamp: 'di 10/23/97 19:58'!
right
	^ right! !

!TextLine methodsFor: 'accessing' stamp: 'di 10/21/97 20:42'!
rightMargin
	"This has to get fixed -- store during composition"
	^ self right! !

!TextLine methodsFor: 'accessing' stamp: 'di 11/26/97 16:18'!
setRight: x
	right _ x! !

!TextLine methodsFor: 'updating' stamp: 'di 10/20/97 23:25'!
slide: delta 
	"Change the starting and stopping points of the line by delta."

	firstIndex _ firstIndex + delta.
	lastIndex _ lastIndex + delta! !

!TextLine methodsFor: 'updating' stamp: 'di 4/28/1999 11:12'!
slideIndexBy: delta andMoveTopTo: newTop
	"Relocate my character indices and y-values.
	Used to slide constant text up or down in the wake of a text replacement."

	firstIndex _ firstIndex + delta.
	lastIndex _ lastIndex + delta.
	bottom _ bottom + (newTop - top).
	top _ newTop.
! !

!TextLine methodsFor: 'accessing' stamp: 'di 10/20/97 23:27'!
stop: stopInteger 
	"Set the stopping point in the string of the line to be stopInteger."

	lastIndex _ stopInteger! !

!TextLine methodsFor: 'accessing' stamp: 'di 10/23/97 19:58'!
top
	^ top! !

!TextLine methodsFor: 'accessing' stamp: 'di 10/23/97 19:58'!
topLeft
	^ left @ top! !

!TextLine methodsFor: 'accessing' stamp: 'di 11/26/97 16:58'!
width
	^ right - left! !

TextMorph subclass: #StringMorphEditor
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Text Support'!

!StringMorphEditor commentStamp: '<historical>' prior: 0!
I am a textMorph used as a pop-up editor for StringMorphs.  I present a yellow background and I go away when a CR is typed or when the user clicks elsewhere.!

!StringMorphEditor class methodsFor: 'new-morph participation' stamp: 'kfr 5/1/2000 13:41'!
includeInNewMorphMenu
	"Not to be instantiated from the menu"
	^ false! !

!StringMorphEditor methodsFor: 'drawing' stamp: 'sw 9/7/1999 16:22'!
drawOn: aCanvas

	aCanvas fillRectangle: self bounds color: Color yellow muchLighter.
	^ super drawOn: aCanvas! !

!StringMorphEditor methodsFor: 'display' stamp: 'sw 4/20/2003 15:46'!
initialize
	"Initialize the receiver.  Give it a white background"

	super initialize.
	self backgroundColor: Color white.
	self color: Color red! !

!StringMorphEditor methodsFor: 'event handling' stamp: 'nk 6/12/2004 22:07'!
keyStroke: evt
	"This is hugely inefficient, but it seems to work, and it's unlikely it will ever need
	to be any more efficient -- it's only intended to edit single-line strings."

	| char priorEditor newSel |
	(((char _ evt keyCharacter) = Character enter) or: [(char = Character cr)
			or: [char = $s and: [evt commandKeyPressed]]])
				ifTrue: [owner doneWithEdits; acceptContents.
	self flag: #arNote. "Probably unnecessary"
						evt hand releaseKeyboardFocus.
						^ self delete].
	
	(char = $l and: [evt commandKeyPressed]) ifTrue:   "cancel"
		[owner cancelEdits.
		evt hand releaseKeyboardFocus.
		^ self delete].

	super keyStroke: evt.
	owner interimContents: self contents asString.
	newSel _ self editor selectionInterval.

	priorEditor _ self editor.  "Save editor state"
	self releaseParagraph.  "Release paragraph so it will grow with selection."
	self paragraph.      "Re-instantiate to set new bounds"
	self installEditorToReplace: priorEditor.  "restore editor state"
	self editor selectFrom: newSel first to: newSel last.
! !

!StringMorphEditor methodsFor: 'event handling' stamp: 'nk 1/23/2004 13:18'!
keyboardFocusChange: aBoolean
	| hadFocus |
	owner ifNil: [ ^self ].
	hadFocus _ owner hasFocus.
	super keyboardFocusChange: aBoolean.
	aBoolean ifFalse:
		[hadFocus ifTrue:
			[owner lostFocusWithoutAccepting; doneWithEdits].
		^ self delete]! !

TextMorph subclass: #TextMorphForEditView
	instanceVariableNames: 'editView acceptOnCR'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Text Support'!

!TextMorphForEditView class methodsFor: 'new-morph participation' stamp: 'kfr 5/1/2000 13:41'!
includeInNewMorphMenu
	"Not to be instantiated from the menu"
	^ false! !

!TextMorphForEditView methodsFor: 'editing' stamp: 'di 4/22/1998 10:57'!
acceptContents
	"The message is sent when the user hits enter or Cmd-S.
	Accept the current contents and end editing."
	self updateFromParagraph.
	editView accept.! !

!TextMorphForEditView methodsFor: 'editing' stamp: 'sw 8/12/2002 00:02'!
acceptOnCR
	"Answer whether the receiver wants to accept when the Return key is hit"

	^ acceptOnCR == true! !

!TextMorphForEditView methodsFor: 'accept/cancel' stamp: 'di 9/11/1998 15:42'!
acceptOnCR: trueOrFalse
	acceptOnCR _ trueOrFalse! !

!TextMorphForEditView methodsFor: 'event handling' stamp: 'ar 11/8/2000 15:50'!
autoScrollView: evt
	"This is kind of a hack because the PluggableTextMorph expects me to first expand the selection before auto scrolling will work."
	| localEvt |
	localEvt _ evt transformedBy: (self transformedFrom: editView).
	super mouseMove: localEvt.
	editView scrollSelectionIntoView: localEvt.! !

!TextMorphForEditView methodsFor: 'editing' stamp: 'di 4/22/1998 11:03'!
cancelEdits
	"The message is sent when the user hits enter or Cmd-L.
	Cancel the current contents and end editing."
	self releaseParagraph.
	editView cancel! !

!TextMorphForEditView methodsFor: 'objects from disk' stamp: 'RAA 12/20/2000 17:50'!
convertToCurrentVersion: varDict refStream: smartRefStrm
	
	acceptOnCR ifNil: [acceptOnCR _ false].
	^super convertToCurrentVersion: varDict refStream: smartRefStrm.

! !

!TextMorphForEditView methodsFor: 'drawing' stamp: 'di 6/22/1998 10:44'!
drawNullTextOn: aCanvas
	"Just run the normal code to show selection in a window"
	aCanvas paragraph: self paragraph bounds: bounds color: color
! !

!TextMorphForEditView methodsFor: 'edit view' stamp: 'di 6/22/1998 01:31'!
editView
	^ editView! !

!TextMorphForEditView methodsFor: 'macpal' stamp: 'di 11/10/1998 10:13'!
flash
	^ editView flash! !

!TextMorphForEditView methodsFor: 'editing' stamp: 'di 10/5/1998 14:03'!
handleInteraction: interActionBlock fromEvent: evt
	"Overridden to pass along a model to the editor for, eg, link resolution, doits, etc"

	self editor model: editView model.  "For evaluateSelection, etc"
	^ super handleInteraction: interActionBlock fromEvent: evt! !

!TextMorphForEditView methodsFor: 'editing' stamp: 'di 4/21/1998 13:23'!
hasUnacceptedEdits: aBoolean
	"Set the hasUnacceptedEdits flag in my view."

	editView hasUnacceptedEdits: aBoolean! !

!TextMorphForEditView methodsFor: 'initialization' stamp: 'di 9/11/1998 15:43'!
initialize
	super initialize.
	acceptOnCR _ false! !

!TextMorphForEditView methodsFor: 'event handling' stamp: 'di 6/30/1998 08:50'!
mouseDown: event

	event yellowButtonPressed ifTrue: [^ editView yellowButtonActivity: event shiftPressed].
	^ super mouseDown: event
! !

!TextMorphForEditView methodsFor: 'event handling' stamp: 'ar 9/26/2001 22:28'!
mouseMove: evt
	| editEvt |
	super mouseMove: evt.
	evt redButtonPressed ifFalse: [^ self].
	editEvt _ evt transformedBy: (self transformedFrom: editView) inverseTransformation.
	(editEvt position y between: editView top and: editView bottom) ifFalse:[
		"Start auto-scrolling"
		self startStepping: #autoScrollView:
			at: Time millisecondClockValue
			arguments: (Array with: editEvt)
			stepTime: 100. "fast enough"
	] ifTrue:[
		self stopSteppingSelector: #autoScrollView:.
	].! !

!TextMorphForEditView methodsFor: 'event handling' stamp: 'fbs 1/7/2005 15:43'!
preferredKeyboardPosition

	| pos |
	pos _ super preferredKeyboardPosition.
	^ pos + (self bounds: self bounds in: World) topLeft.
! !

!TextMorphForEditView methodsFor: 'miscellaneous' stamp: 'sw 7/27/2001 13:35'!
selectAll
	"Tell my editor to select all the text"

	self editor selectAll! !

!TextMorphForEditView methodsFor: 'edit view' stamp: 'di 4/21/1998 13:09'!
setEditView: editPane
	editView _ editPane! !

!TextMorphForEditView methodsFor: 'debug and other' stamp: 'sw 11/2/1998 15:51'!
tempCommand
	"Smalltalk browseAllImplementorsOf: #tempCommand"
	"Place your definition for tempCommand for this class here"! !

!TextMorphForEditView methodsFor: 'private' stamp: 'dew 2/21/1999 03:09'!
updateFromParagraph  
	super updateFromParagraph.
	editView setScrollDeltas.! !

!TextMorphForEditView methodsFor: 'event handling' stamp: 'sw 8/29/2000 15:06'!
wouldAcceptKeyboardFocusUponTab
	"Answer whether the receiver would be a happy inheritor of keyboard focus if tab were hit in an enclosing playfield under propitious circumstances.  Does not make sense for this kind of morph, which is encased in a window"

	^ false! !

RectangleMorph subclass: #TextFieldMorph
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Text Support'!

!TextFieldMorph commentStamp: '<historical>' prior: 0!
Act as a field in a HyperCard-like setting.  Has both properties of a Rectangle, and exposes some proteries of the TextMorph it owns.

!

!TextFieldMorph class methodsFor: 'scripting' stamp: 'md 11/14/2003 17:32'!
authoringPrototype 
	"Answer an instance of the receiver that can serve as a prototype for authoring"

	| proto |
	proto _ super authoringPrototype.
	proto setProperty: #shared toValue: true.
	proto extent: 170 @ 30.
	proto color: Color veryLightGray lighter.
	proto contents: 'on a clear day you can...'.
	^ proto
! !

!TextFieldMorph class methodsFor: 'class initialization' stamp: 'asm 4/11/2003 11:56'!
initialize

	self registerInFlapsRegistry.	! !

!TextFieldMorph class methodsFor: 'class initialization' stamp: 'asm 4/11/2003 11:58'!
registerInFlapsRegistry
	"Register the receiver in the system's flaps registry"
	self environment
		at: #Flaps
		ifPresent: [:cl | cl registerQuad: #(TextFieldMorph  exampleBackgroundField	'Scrolling Field'	'A scrolling data field which will have a different value on every card of the background')
						forFlapNamed: 'Scripting'.]! !

!TextFieldMorph class methodsFor: 'class initialization' stamp: 'asm 4/11/2003 12:41'!
unload
	"Unload the receiver from global registries"

	self environment at: #Flaps ifPresent: [:cl |
	cl unregisterQuadsWithReceiver: self] ! !

!TextFieldMorph methodsFor: 'just like textMorph' stamp: 'tk 9/6/2000 11:03'!
append: stringOrText
	"add to my text"
	| tm |

	(tm _ self findA: TextMorph) ifNil: [^ nil].
	tm contents append: stringOrText.
	tm releaseParagraph; paragraph.


	! !

!TextFieldMorph methodsFor: 'just like textMorph' stamp: 'tk 8/30/2000 14:22'!
contents
	| tm |
	"talk to my text"

	(tm _ self findA: TextMorph) ifNil: [^ nil].
	^ tm contents! !

!TextFieldMorph methodsFor: 'just like textMorph' stamp: 'ar 4/10/2005 18:54'!
contents: textOrString
	"talk to my text"
	| tm newText atts |

	(tm _ self findA: TextMorph) ifNil: [^ nil].
	textOrString isString ifTrue: [
		tm contents ifNotNil: ["Keep previous properties of the field"
			newText _ textOrString asText.
			atts _ tm contents attributesAt: 1.
			atts do: [:each | newText addAttribute: each].
			^ tm contents: newText]].

	^ tm contents: textOrString! !

!TextFieldMorph methodsFor: 'initialization' stamp: 'dgd 3/7/2003 14:54'!
defaultColor
"answer the default color/fill style for the receiver"
	^ Color veryLightGray lighter! !

!TextFieldMorph methodsFor: 'just like textMorph' stamp: 'tk 9/4/2000 16:28'!
fit
	"tell my text to recompute its looks"
	| tm |

	(tm _ self findA: TextMorph) ifNil: [^ nil].
	tm releaseParagraph; paragraph.! !

!TextFieldMorph methodsFor: 'just like textMorph' stamp: 'tk 8/30/2000 14:24'!
fontName: fontName size: fontSize
	| tm |
	"talk to my text"

	(tm _ self findA: TextMorph) ifNil: [^ nil].
	^ tm fontName: fontName size: fontSize
! !

!TextFieldMorph methodsFor: 'initialization' stamp: 'dgd 3/7/2003 14:54'!
initialize
	"initialize the state of the receiver"
	| tm |
	super initialize.
	""
	
	self addMorph: (tm _ TextMorph new).
	tm fillingOnOff! !

!TextFieldMorph methodsFor: 'just like textMorph' stamp: 'tk 9/6/2000 12:33'!
lineCount
	| tm |
	"how many lines in my text"

	(tm _ self findA: TextMorph) ifNil: [^ nil].
	^ tm contents string lineCount! !

!TextFieldMorph methodsFor: 'just like textMorph' stamp: 'ar 12/27/2001 00:03'!
prepend: stringOrText
	"add to my text"
	| tm |

	(tm _ self findA: TextMorph) ifNil: [^ nil].
	tm contents prepend: stringOrText.
	tm releaseParagraph; paragraph.


	! !

CommandHistory subclass: #TextMorphCommandHistory
	instanceVariableNames: 'textMorph'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Text Support'!

!TextMorphCommandHistory methodsFor: 'command exec' stamp: 'sps 7/23/2003 18:57'!
redo
	^super redoNextCommand
! !

!TextMorphCommandHistory methodsFor: 'command exec' stamp: 'sps 7/23/2003 18:58'!
rememberCommand: aCommand
	"Make the supplied command be the 'LastCommand', and mark it 'done'"

	"Before adding the new command, remove any commands after the last #done 
	command, and make that last #done command be lastCommand."
	self removeUndoneCommands.
	aCommand phase: #done.
		
	"If we are building a compound command, just add the new command to that"
	history addLast: aCommand.
	lastCommand _ aCommand.
"Debug dShow: ('Remember: ', commandToUse asString)."

! !

!TextMorphCommandHistory methodsFor: 'command exec' stamp: 'sps 7/24/2003 16:42'!
removeUndoneCommands
"Remove all of the commands at the end of history until the first one that is not marked #undone"

	history reversed do: [ :command |
		(command phase == #done) ifTrue:[
			lastCommand _ command.
			^self
		]ifFalse:[
			history remove: command.
		].
	].
	
	"If there were no #done commands on the stack, then get rid of lastCommand"
	lastCommand _ nil.
! !

!TextMorphCommandHistory methodsFor: 'command exec' stamp: 'sps 7/23/2003 18:57'!
undo
	^super undoLastCommand

! !

TextAttribute subclass: #TextAnchor
	instanceVariableNames: 'anchoredMorph'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Text Support'!

!TextAnchor commentStamp: 'md 8/10/2006 11:52' prior: 0!
TextAnchors support anchoring of images in text.  A TextAnchor exists as an attribute of text emphasis, and it gets control like a FontReference, through the emphasizeScanner: message.  Depending on whether its anchoredMorph is a Morph or a Form, it repositions the morph, or displays the form respectively.  The coordination between composition, display and selection can best be understood by browsing the various implementations of placeEmbeddedObject:.

In the morphic world, simply embed any form or morph in text.

	Workspace new
		contents: (Text withAll: 'foo') , (Text string: '*' attribute: (TextAnchor new anchoredMorph: MenuIcons confirmIcon)) , (Text withAll: 'bar');
		openLabel: 'Text with Form'.

	Workspace new
		contents: (Text withAll: 'foo') , (Text string: '*' attribute: (TextAnchor new anchoredMorph: EllipseMorph new)) , (Text withAll: 'bar');
		openLabel: 'Text with Morph'.

In this case you select a piece of the screen, and it gets anchored to a one-character text in the editor's past buffer.  If you then paste into some other text, you will see the image as an embedded image.!

!TextAnchor methodsFor: 'comparing' stamp: 'di 7/1/1998 14:35'!
= other 
	^ (other class == self class) 
		and: [other anchoredMorph == anchoredMorph]! !

!TextAnchor methodsFor: 'accessing' stamp: 'di 11/10/97 13:21'!
anchoredMorph
	^ anchoredMorph! !

!TextAnchor methodsFor: 'accessing' stamp: 'di 11/10/97 10:47'!
anchoredMorph: aMorph 
	anchoredMorph _ aMorph! !

!TextAnchor methodsFor: 'accessing' stamp: 'sw 11/9/1999 17:21'!
couldDeriveFromPrettyPrinting
	^ false! !

!TextAnchor methodsFor: 'visiting' stamp: 'lr 2/3/2006 16:13'!
emphasizeScanner: aScanner
	self anchoredMorph ifNil: [ ^ self ].
	aScanner placeEmbeddedObject: self anchoredMorph.! !

!TextAnchor methodsFor: 'comparing' stamp: 'ar 9/9/2003 22:03'!
hash
	"#hash is re-implemented because #= is re-implemented"
	^anchoredMorph identityHash! !

!TextAnchor methodsFor: 'accessing' stamp: 'di 11/10/97 14:08'!
mayBeExtended
	"A textAnchor is designed to modify only a single character, and therefore must not be extended by the ParagraphEditor's emphasisHere facility"
	^ false! !
TextContainer initialize!
TextFieldMorph initialize!
